"""Constants used for internally processing file chunks and for some default parameter values"""

#imports
import re

class DataFileHandlingConstants :
    """
    Constants for internally handling DataFileChunks
    """

    CHUNK_ALREADY_WRITTEN_CODE = 10 # code indicating that a particular chunk has already been written
    FILE_HASH_MISMATCH_CODE = -1 # code indicating that a file's hashes didn't match
    FILE_SUCCESSFULLY_RECONSTRUCTED_CODE = 3 # code indicating that a file was successfully fully reconstructed
    FILE_IN_PROGRESS = 2 # code indicating that a file is in the process of being reconstructed

DATA_FILE_HANDLING_CONST=DataFileHandlingConstants()

class RunOptionConstants :
    """
    Some default command line options/function kwargs
    """

    DEFAULT_CONFIG_FILE = 'test' # name of the config file that will be used by default
    DEFAULT_TOPIC_NAME = 'test' # name of the topic to produce to by default
    PRODUCTION_CONFIG_FILE = 'prod' # name of the config file used in "real" production
    N_DEFAULT_UPLOAD_THREADS = 2 # default number of threads to use when uploading a file
    N_DEFAULT_DOWNLOAD_THREADS = 2 # default number of threads to use when downloading chunks of a file

    #matches everything except something starting with a '.' or ending in '.log'
    DEFAULT_UPLOAD_REGEX = re.compile(r'^((?!(\.|.*.log))).*$')

    DEFAULT_CHUNK_SIZE = 16384 # default size in bytes of each file upload chunk
    DEFAULT_MAX_UPLOAD_QUEUE_SIZE = 3000 # default maximum number of items allowed in the upload Queue at once

RUN_OPT_CONST = RunOptionConstants()



"""Small class to effect transfer of a datafile to an S3 bucket"""

# @author: Amir H. Sharifzadeh, https://pages.jh.edu/asharif9/
# @project: OpenMSI
# @date: 04/12/2022
# @owner: The Institute of Data Intensive Engineering and Science, https://idies.jhu.edu/
# Johns Hopkins University http://www.jhu.edu
# import logging

from botocore.exceptions import ClientError
from .s3_service import S3Service

class S3DataTransfer(S3Service) :
    """
    Small class to effect transfer of a datafile to an S3 bucket
    """

    def __init__(self, s3_config, *args, **kwargs):
        super().__init__(s3_config,*args,**kwargs)

    def transfer_object_stream(self,object_key,datafile):
        """
        Transfer the contents of a DataFile Consumed from chunks in a topic to an S3 bucket
        """
        file_name = str(datafile.filename)
        try:
            self.s3_client.put_object(Body=datafile.bytestring, Bucket=self.bucket_name,
                                      Key=object_key,
                                      # GrantRead=self.grant_read
                                      )
            msg = f'{file_name} successfully transferred into {self.bucket_name}/{object_key}'
            self.logger.info(msg)
        except ClientError as err :
            self.logger.error(f'{err.response}: failed to transfer {file_name} into {self.bucket_name}/{object_key}')



"""Most classes in this directory are imported from the main module, but there are some other utility classes here"""

from .entity.data_file import DataFile
from .entity.download_data_file import DownloadDataFile, DownloadDataFileToDisk, DownloadDataFileToMemory
from .entity.data_file_directory import DataFileDirectory
from .entity.reproducer_message import ReproducerMessage

__all__ = [
    'DataFile',
    'DownloadDataFile',
    'DownloadDataFileToDisk',
    'DownloadDataFileToMemory',
    'DataFileDirectory',
    'ReproducerMessage',
]



"""Script to invoke the OS-appropriate ServiceManager class to do anything other than installing"""

#imports
from .config import SERVICE_CONST
from .utilities import get_os_name
from .service_manager_base import ServiceManagerBase
from .windows_service_manager import WindowsServiceManager
from .linux_service_manager import LinuxServiceManager

def main(given_args=None) :
    """
    Use the OS-appropriate ServiceManager class to manage a Service/daemon (i.e., anything except installing it)
    """
    #get the arguments
    parser = ServiceManagerBase.get_argument_parser('manage')
    args = parser.parse_args() if given_args is None else parser.parse_args(given_args)
    #get the name of the OS and start the object
    operating_system = get_os_name()
    manager_args = [args.service_name]
    manager_kwargs = {'interactive':given_args is None,
                      'logger':SERVICE_CONST.logger}
    managers_by_os_name = {'Windows':WindowsServiceManager,
                           'Linux':LinuxServiceManager,}
    manager_class = managers_by_os_name[operating_system]
    manager = manager_class(*manager_args,**manager_kwargs)
    #run some function based on the run mode
    manager.run_manage_command(args.run_mode,
                               remove_env_vars=args.remove_env_vars,
                               remove_install_args=args.remove_install_args,
                               remove_nssm=args.remove_nssm)

if __name__=='__main__' :
    main()



"""A set of Producers sharing an optional KafkaCrypto instance for key passing"""

#imports
from ..utilities import LogOwner
from .openmsistream_producer import OpenMSIStreamProducer

class ProducerGroup(LogOwner) :
    """
    Class for working with a group of producers sharing a single :class:`kafkacrypto.KafkaCrypto` instance

    :param config_path: Path to the config file that should be used to define Producers in the group
    :type config_path: :class:`pathlib.Path`
    :param kafkacrypto: The :class:`~OpenMSIStreamKafkaCrypto` object that should be used to instantiate Producers.
        Only needed if a single specific :class:`~OpenMSIStreamKafkaCrypto` instance should be shared.
    :type kafkacrypto: :class:`~OpenMSIStreamKafkaCrypto`, optional
    """

    @property
    def kafkacrypto(self) :
        """
        The KafkaCrypto object handling key passing and serialization for encrypted messages
        """
        return self.__p_kwargs['kafkacrypto'] if 'kafkacrypto' in self.__p_kwargs else None

    def __init__(self,config_path,kafkacrypto=None,**kwargs) :
        """
        Constructor method
        """
        super().__init__(**kwargs)
        self.__p_args, self.__p_kwargs = OpenMSIStreamProducer.get_producer_args_kwargs(config_path,
                                                                                        logger=self.logger,
                                                                                        kafkacrypto=kafkacrypto)

    def get_new_producer(self) :
        """
        Return a new :class:`~OpenMSIStreamProducer` object.
        Call this function from a child thread to get thread-independent Producers.
        Note: this function just creates the Producer; closing it etc. must be handled by whatever calls this function.

        :return: a Producer created using the config set in the constructor
        :rtype: :class:`~OpenMSIStreamProducer`
        """
        producer = OpenMSIStreamProducer(*self.__p_args,**self.__p_kwargs)
        return producer

    def close(self) :
        """
        Wrapper around :func:`kafkacrypto.KafkaCrypto.close`.
        """
        try :
            self.__p_kwargs['kafkacrypto'].close()
        except Exception :
            pass
        finally :
            self.__p_kwargs['kafkacrypto'] = None



"""
Code for managing the two .csv files listing files that are
being reconstructed from the topic and files that have been fully handled
"""

#imports
import datetime, re
from abc import ABC
from dataclasses import dataclass
from ....utilities import DataclassTable, LogOwner
from ...utilities import get_message_prepend

@dataclass
class StreamHandlerRegistryLineInProgress :
    """
    A line in the table listing files in progress
    """
    filename : str #the name of the file
    rel_filepath : str #the (posix) path to the file, relative to its root directory
    status : str #the status of the file, either "in_progress", "failed", or "mismatched_hash"
    n_chunks : int #the total number of chunks in the file
    first_message : datetime.datetime #timestamp of when the first message for the file was consumed
    most_recent_message : datetime.datetime #timestamp of when the most recent chunk from the file was consumed

@dataclass
class StreamHandlerRegistryLineSucceeded :
    """
    A line in the table listing files that have been successfully handled
    """
    filename : str #the name of the file
    rel_filepath : str #the (posix) path to the file, relative to its root directory
    n_chunks : int #the total number of chunks in the file
    first_message : datetime.datetime #timestamp of when the most recent chunk from the file was consumed
    processed_at : datetime.datetime #timestamp of when the file was processed

class StreamHandlerRegistry(LogOwner,ABC) :
    """
    A general base class to keep track of the status of files read during stream handling
    """

    IN_PROGRESS = 'in_progress'
    MISMATCHED_HASH = 'mismatched_hash'

    @property
    def in_progress_table(self) :
        """
        The table listing files in progress
        """
        return self._in_progress_table

    @property
    def succeeded_table(self) :
        """
        The table listing files that have been successfully handled
        """
        return self._succeeded_table

    @property
    def filepaths_to_rerun(self) :
        """
        A list of all the filepaths that have not yet been successfully processed
        """
        to_rerun = []
        for addr in self._in_progress_table.obj_addresses :
            to_rerun.append(self._in_progress_table.get_entry_attrs(addr,'rel_filepath'))
        return to_rerun

    @property
    def rerun_file_key_regex(self) :
        """
        Regex matching keys for messages from files that need to be rerun
        """
        if self.n_files_to_rerun<=0 :
            return None
        regex_str = r'^('
        for fp in self.filepaths_to_rerun :
            if r'/' in fp :
                subdir_str = f'{"/".join((fp.split("/"))[:-1])}'
                filename = (fp.split("/"))[-1]
            else :
                subdir_str = None
                filename = fp
            regex_str+=f'{get_message_prepend(subdir_str,filename)}|'
        regex_str=f'{regex_str[:-1]})_.*$'
        return re.compile(regex_str)

    @property
    def n_files_to_rerun(self) :
        """
        The number of files in the registry that are marked as "in_progress"
        """
        return len(self.filepaths_to_rerun)

    def __init__(self,in_progress_filepath,succeeded_filepath,*args,**kwargs) :
        """
        in_progress_filepath = path to the file that should hold the "in_progress" datatable
        succeeded_filepath = path to the file that should hole the "succeeded" datatable
        """
        super().__init__(*args,**kwargs)
        self._in_progress_table = DataclassTable(dataclass_type=StreamHandlerRegistryLineInProgress,
                                                  filepath=in_progress_filepath,logger=self.logger)
        self._succeeded_table = DataclassTable(dataclass_type=StreamHandlerRegistryLineSucceeded,
                                                  filepath=succeeded_filepath,logger=self.logger)

    def register_file_in_progress(self,dfc) :
        """
        Add/update a line in the table showing that a file is in progress
        """
        self._add_or_modify_in_progress_entry(dfc,self.IN_PROGRESS)

    def register_file_mismatched_hash(self,dfc) :
        """
        Add/update a line in the table to show that a consumed file was mismatched with its original content hash
        """
        self._add_or_modify_in_progress_entry(dfc,self.MISMATCHED_HASH)

    def _add_or_modify_in_progress_entry(self,dfc,new_status) :
        filename, rel_filepath = self._get_name_and_rel_filepath_for_data_file_chunk(dfc)
        existing_entry_addr = self._get_in_progress_address_for_rel_filepath(rel_filepath)
        if existing_entry_addr is None :
            new_entry = StreamHandlerRegistryLineInProgress(filename,
                                                              rel_filepath,
                                                              new_status,
                                                              dfc.n_total_chunks,
                                                              datetime.datetime.now(),
                                                              datetime.datetime.now())
            self._in_progress_table.add_entries(new_entry)
            self._in_progress_table.dump_to_file()
        else :
            self._in_progress_table.set_entry_attrs(existing_entry_addr,
                                                    status=new_status,
                                                    most_recent_message=datetime.datetime.now())

    def _get_name_and_rel_filepath_for_data_file_chunk(self,dfc) :
        filename = dfc.filename
        rel_filepath = f'{dfc.subdir_str}/{filename}' if dfc.subdir_str!='' else filename
        return filename,rel_filepath

    def _get_in_progress_address_for_rel_filepath(self,rel_filepath) :
        existing_obj_addresses = self._in_progress_table.obj_addresses_by_key_attr('rel_filepath')
        if rel_filepath not in existing_obj_addresses :
            return None
        if len(existing_obj_addresses[rel_filepath])!=1 :
            errmsg = f'ERROR: found more than one {self.__class__.__name__} entry for relative filepath '
            errmsg+= f'{rel_filepath} in file at {self._in_progress_table.filepath}'
            self.logger.error(errmsg,ValueError)
        return existing_obj_addresses[rel_filepath][0]

class StreamProcessorRegistry(StreamHandlerRegistry) :
    """
    A class to keep track of the status of files read during stream processing
    """

    FAILED = 'failed'

    def __init__(self,dirpath,topic_name,consumer_group_id,*args,**kwargs) :
        """
        dirpath    = path to the directory that should contain the csv files
        topic_name = the name of the topic that will be produced to (used in the filenames)
        """
        in_progress_filepath = dirpath / f'files_consumed_from_{topic_name}_by_{consumer_group_id}.csv'
        succeeded_filepath = dirpath / f'files_successfully_processed_from_{topic_name}_by_{consumer_group_id}.csv'
        super().__init__(in_progress_filepath,succeeded_filepath,*args,**kwargs)

    def register_file_successfully_processed(self,dfc) :
        """
        Add/update a line in the table to show that a file has successfully been processed
        """
        filename, rel_filepath = self._get_name_and_rel_filepath_for_data_file_chunk(dfc)
        existing_entry_addr = self._get_in_progress_address_for_rel_filepath(rel_filepath)
        if existing_entry_addr is not None :
            attrs = self._in_progress_table.get_entry_attrs(existing_entry_addr)
            new_entry = StreamHandlerRegistryLineSucceeded(attrs['filename'],
                                                           attrs['rel_filepath'],
                                                           attrs['n_chunks'],
                                                           attrs['first_message'],
                                                           datetime.datetime.now())
        else :
            new_entry = StreamHandlerRegistryLineSucceeded(filename,
                                                           rel_filepath,
                                                           dfc.n_total_chunks,
                                                           datetime.datetime.now(),
                                                           datetime.datetime.now())
        self._succeeded_table.add_entries(new_entry)
        self._succeeded_table.dump_to_file()
        if existing_entry_addr is not None :
            self._in_progress_table.remove_entries(existing_entry_addr)
            self._in_progress_table.dump_to_file()

    def register_file_processing_failed(self,dfc) :
        """
        Add/update a line in the table to show that a file has failed to be processed
        """
        self._add_or_modify_in_progress_entry(dfc,self.FAILED)

class StreamReproducerRegistry(StreamHandlerRegistry) :
    """
    A class to keep track of the status of files read from one topic with associated information
    created and re-produced to a different topic
    """

    PRODUCING_MESSAGE_FAILED = 'producing_message_failed'
    COMPUTING_RESULT_FAILED = 'computing_result_message_failed'

    def __init__(self,dirpath,consumer_topic_name,producer_topic_name,consumer_group_id,*args,**kwargs) :
        """
        dirpath    = path to the directory that should contain the csv file
        topic_name = the name of the topic that will be produced to (used in the filename)
        """
        in_progress_filepath = dirpath / f'files_consumed_from_{consumer_topic_name}_by_{consumer_group_id}.csv'
        succeeded_filepath = dirpath / f'files_with_results_produced_to_{producer_topic_name}.csv'
        super().__init__(in_progress_filepath,succeeded_filepath,*args,**kwargs)

    def register_file_results_produced(self,filename,rel_filepath,n_total_chunks) :
        """
        Add/update a line in the table to show that a file has successfully been processed
        """
        existing_entry_addr = self._get_in_progress_address_for_rel_filepath(rel_filepath)
        if existing_entry_addr is not None :
            attrs = self._in_progress_table.get_entry_attrs(existing_entry_addr)
            new_entry = StreamHandlerRegistryLineSucceeded(attrs['filename'],
                                                           attrs['rel_filepath'],
                                                           attrs['n_chunks'],
                                                           attrs['first_message'],
                                                           datetime.datetime.now())
        else :
            new_entry = StreamHandlerRegistryLineSucceeded(filename,
                                                           rel_filepath,
                                                           n_total_chunks,
                                                           datetime.datetime.now(),
                                                           datetime.datetime.now())
        self._succeeded_table.add_entries(new_entry)
        self._succeeded_table.dump_to_file()
        if existing_entry_addr is not None :
            self._in_progress_table.remove_entries(existing_entry_addr)
            self._in_progress_table.dump_to_file()

    def register_file_computing_result_failed(self,filename,rel_filepath,n_total_chunks) :
        """
        Update a line in the table to show that a file failed to have its processing result computed
        """
        self._add_or_modify_in_progress_entry_without_chunk(filename,rel_filepath,n_total_chunks,
                                                            self.COMPUTING_RESULT_FAILED)

    def register_file_result_production_failed(self,filename,rel_filepath,n_total_chunks) :
        """
        Update a line in the table to show that the call to Producer.produce() for a message computed
        from a file returned an error
        """
        self._add_or_modify_in_progress_entry_without_chunk(filename,rel_filepath,n_total_chunks,
                                                            self.PRODUCING_MESSAGE_FAILED)

    def _add_or_modify_in_progress_entry_without_chunk(self,filename,rel_filepath,n_total_chunks,new_status) :
        existing_entry_addr = self._get_in_progress_address_for_rel_filepath(rel_filepath)
        if existing_entry_addr is None :
            new_entry = StreamHandlerRegistryLineInProgress(filename,
                                                              rel_filepath,
                                                              new_status,
                                                              n_total_chunks,
                                                              datetime.datetime.now(),
                                                              datetime.datetime.now())
            self._in_progress_table.add_entries(new_entry)
            self._in_progress_table.dump_to_file()
        else :
            self._in_progress_table.set_entry_attrs(existing_entry_addr,
                                                    status=new_status,
                                                    most_recent_message=datetime.datetime.now())



"""Classes for serialization and deserialization operations"""

#imports
import pathlib, inspect, time, warnings
from hashlib import sha512
import msgpack
from confluent_kafka.error import SerializationError
from confluent_kafka.serialization import Serializer, Deserializer
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto.message import KafkaCryptoMessage, KafkaCryptoMessageError
from ..data_file_io.entity.data_file_chunk import DataFileChunk

####################### COMPOUND (DE)SERIALIZERS FOR STACKING MULTIPLE STEPS #######################

class CompoundSerDes(Serializer, Deserializer):
    """
    A Serializer/Deserializer that stacks multiple Serialization/Deserialization steps
    """

    def __init__(self, *args):
        """
        args = a list of (De)Serializer (or otherwise callable) objects to apply IN GIVEN ORDER to some data
        """
        self.__steps = list(args)

    def __call__(self,data,ctx=None) :
        if data is None :
            return None
        for istep,serdes in enumerate(self.__steps,start=1) :
            try :
                data = serdes(data)
            except Exception as exc :
                errmsg = f'ERROR: failed to (de)serialize at step {istep} (of {len(self.__steps)}) in CompoundSerDes! '
                errmsg+= f'Callable = {serdes}, exception = {exc}'
                raise SerializationError(errmsg)
        return data

class CompoundSerializer(Serializer):
    """
    A Serializer that stacks multiple steps
    For use with KafkaCrypto since topic names must be passed
    """

    def __init__(self, *args):
        self.__steps = list(args)

    def serialize(self,topic,data) :
        """
        Call a chain of Serializers for a given message (may contain encryption)
        """
        for istep,ser in enumerate(self.__steps,start=1) :
            try :
                if hasattr(ser,'serialize') and inspect.isroutine(ser.serialize) :
                    data = ser.serialize(topic,data)
                else :
                    data = ser(data,ctx=None)
            except Exception as exc :
                errmsg = f'ERROR: failed to serialize at step {istep} (of {len(self.__steps)}) in '
                errmsg+= f'{self.__class__.__name__}! Callable = {ser}, exception = {exc}'
                raise SerializationError(errmsg)
        return data

    def __call__(self,data,ctx=None) :
        if data is None :
            return None
        raise SerializationError(f'ERROR: __call__ method not implemented for a {self.__class__.__name__}')

class CompoundDeserializer(Deserializer):
    """
    A Deserializer that stacks multiple steps
    For use with KafkaCrypto since topic names must be passed
    """

    MAX_WAIT_PER_DECRYPT = 60.0 #in seconds

    def __init__(self, *args):
        self.__steps = list(args)

    def deserialize(self,topic,data) :
        """
        Call a chain of Deserializers for a given message (may contain decryption)
        """
        for istep,des in enumerate(self.__steps,start=1) :
            try :
                data = self.__deserialize_message(des,topic,data)
            except Exception as exc :
                errmsg = f'ERROR: failed to deserialize at step {istep} (of {len(self.__steps)})'
                errmsg+= f' in {self.__class__.__name__}! Callable = {des}, exception = {exc}'
                raise SerializationError(errmsg)
        return data

    def __call__(self,data,ctx=None) :
        if data is None :
            return None
        raise SerializationError(f'ERROR: __call__ method not implemented for a {self.__class__.__name__}')

    def __deserialize_message(self,des,topic,data) :
        """
        Deserialize a (possibly encrypted) message
        """
        #deserialize a KafkaCrypto message
        if hasattr(des,'deserialize') and inspect.isroutine(des.deserialize) :
            data = des.deserialize(topic,data)
            if isinstance(data,KafkaCryptoMessage) :
                success = False
                elapsed = 0
                while (not success) and elapsed<self.MAX_WAIT_PER_DECRYPT :
                    try :
                        data = data.getMessage()
                        success = True
                    except KafkaCryptoMessageError :
                        time.sleep(0.1)
                        elapsed+=0.1
                #if the message could not be decrypted, return the Message object itself
                if not success :
                    return data
        #if a previous step failed to deserialize a KafkaCrypto message, don't bother trying this next step
        elif not isinstance(data,KafkaCryptoMessage) :
            data = des(data,ctx=None)
        return data


####################### SERIALIZING/DESERIALIZING FILE CHUNKS #######################

class DataFileChunkSerializer(Serializer) :
    """
    Serializes DataFileChunk objects to bytesarrays
    """

    def __call__(self,file_chunk_obj,ctx=None) :
        if file_chunk_obj is None :
            return None
        if not isinstance(file_chunk_obj,DataFileChunk) :
            raise SerializationError('ERROR: object passed to FileChunkSerializer is not a DataFileChunk!')
        #pack up all the relevant bits of information into a single bytearray
        try :
            #get the chunk's data from the file if necessary
            if file_chunk_obj.data is None :
                file_chunk_obj.populate_with_file_data()
            #pack up all of the properties of the message
            ordered_properties = []
            ordered_properties.append(str(file_chunk_obj.filename))
            ordered_properties.append(file_chunk_obj.file_hash)
            ordered_properties.append(file_chunk_obj.chunk_hash)
            ordered_properties.append(file_chunk_obj.chunk_offset_write)
            ordered_properties.append(file_chunk_obj.chunk_i)
            ordered_properties.append(file_chunk_obj.n_total_chunks)
            ordered_properties.append(file_chunk_obj.subdir_str)
            ordered_properties.append(file_chunk_obj.filename_append)
            ordered_properties.append(file_chunk_obj.data)
            return msgpack.packb(ordered_properties,use_bin_type=True)
        except Exception as exc :
            raise SerializationError(f'ERROR: failed to serialize a DataFileChunk! Exception: {exc}')

class DataFileChunkDeserializer(Deserializer) :
    """
    Deseralizes bytearrays to DataFileChunk objects
    """

    def __call__(self,byte_array,ctx=None) :
        if byte_array is None :
            return None
        try :
            #unpack the byte array
            ordered_properties = msgpack.unpackb(byte_array,raw=True)
            if len(ordered_properties)!=9 :
                errmsg = 'ERROR: unrecognized token passed to DataFileChunkDeserializer. Expected 9 properties'
                errmsg+= f' but found {len(ordered_properties)}'
                raise ValueError(errmsg)
            try :
                filename = str(ordered_properties[0].decode())
                file_hash = ordered_properties[1]
                chunk_hash = ordered_properties[2]
                chunk_offset_read = None
                chunk_offset_write = int(ordered_properties[3])
                chunk_i = int(ordered_properties[4])
                n_total_chunks = int(ordered_properties[5])
                subdir_str = str(ordered_properties[6].decode())
                filename_append = str(ordered_properties[7].decode())
                data = ordered_properties[8]
            except Exception as exc :
                errmsg = f'ERROR: unrecognized value(s) when deserializing a DataFileChunk from token. Exception: {exc}'
                raise ValueError(errmsg)
            #make sure the hash of the chunk's data matches with what it was before
            check_chunk_hash = sha512()
            check_chunk_hash.update(data)
            check_chunk_hash = check_chunk_hash.digest()
            if check_chunk_hash!=chunk_hash :
                errmsg = f'ERROR: chunk hash {check_chunk_hash} != expected hash {chunk_hash} in file {filename}, '
                errmsg+= f'offset {chunk_offset_write}'
                raise RuntimeError(errmsg)
            #set the filepath based on the subdirectory string
            if subdir_str=='' :
                filepath = pathlib.Path(filename)
            subdir_path = pathlib.PurePosixPath(subdir_str)
            filepath = pathlib.Path('').joinpath(*(subdir_path.parts),filename)
            return DataFileChunk(filepath,filename,file_hash,chunk_hash,chunk_offset_read,chunk_offset_write,
                                 len(data),chunk_i,n_total_chunks,data=data,filename_append=filename_append)
        except Exception as exc :
            raise SerializationError(f'ERROR: failed to deserialize a DataFileChunk! Exception: {exc}')



"""A set of Consumers and Producers that share share the same KafkaCrypto key-passing instance"""

#imports
from ..utilities.logging import LogOwner
from .consumer_group import ConsumerGroup
from .producer_group import ProducerGroup

class ConsumerAndProducerGroup(LogOwner) :
    """
    Class for working with a group of Consumers and Producers sharing a single :class:`kafkacrypto.KafkaCrypto` instance

    :param config_path: Path to the config file that should be used to define Consumers/Producers in the group
    :type config_path: :class:`pathlib.Path`
    :param consumer_topic_name: The name of the topic to which the Consumers should be subscribed
    :type consumer_topic_name: str
    :param consumer_group_id: The ID string that should be used for each Consumer in the group.
        "create_new" (the defaults) will create a new UID to use.
    :type consumer_group_id: str, optional
    """

    @property
    def consumer_topic_name(self) :
        """
        Name of the topic to which the consumers are subscribed
        """
        return self.__consumer_group.topic_name
    @property
    def consumer_group_id(self) :
        """
        Group ID of all consumers in the group
        """
        return self.__consumer_group.consumer_group_id

    def __init__(self,config_path,consumer_topic_name,*,consumer_group_id='create_new',**kwargs) :
        """
        Constructor method
        """
        super().__init__(**kwargs)
        self.__consumer_group = ConsumerGroup(config_path,consumer_topic_name,
                                              consumer_group_id=consumer_group_id,logger=self.logger)
        self.__producer_group = ProducerGroup(config_path,kafkacrypto=self.__consumer_group.kafkacrypto,
                                              logger=self.logger)

    def get_new_subscribed_consumer(self,*,restart_at_beginning=False,**kwargs) :
        """
        Return a new Consumer, subscribed to the topic and with the shared group ID.
        Call this function from a child thread to get thread-independent Consumers.

        Note: This function just creates and subscribes the Consumer. Polling it, closing
        it, and everything else must be handled by whatever calls this function.

        :param restart_at_beginning: if True, the new Consumer will start reading partitions from the earliest
            messages available, regardless of Consumer group ID and auto.offset.reset values.
            Useful when re-reading messages.
        :type restart_at_beginning: bool, optional
        :param kwargs: other keyword arguments are passed to the :class:`~OpenMSIStreamConsumer` constructor method
            (for example, parameters to set a key regex or how to filter messages)
        :type kwargs: dict

        :return: a Consumer created using the configs set in the constructor/from `kwargs`, subscribed to the topic
        :rtype: :class:`~OpenMSIStreamConsumer`
        """
        return self.__consumer_group.get_new_subscribed_consumer(restart_at_beginning=restart_at_beginning,**kwargs)

    def get_new_producer(self) :
        """
        Return a new :class:`~OpenMSIStreamProducer` object.
        Call this function from a child thread to get thread-independent Producers.
        Note: this function just creates the Producer; closing it etc. must be handled by whatever calls this function.

        :return: a Producer created using the config set in the constructor
        :rtype: :class:`~OpenMSIStreamProducer`
        """
        return self.__producer_group.get_new_producer()

    def close(self) :
        """
        Wrapper around :func:`kafkacrypto.KafkaCrypto.close`.
        """
        self.__consumer_group.close()
        self.__producer_group.close()



"""A directory of DataFiles"""

#imports
from ...utilities import LogOwner
from ...utilities.misc import populated_kwargs

class DataFileDirectory(LogOwner) :
    """
    Base class representing any directory holding data files

    :param dirpath: path to the directory
    :type dirpath: :class:`pathlib.Path`
    """

    def __init__(self,dirpath,*args,**kwargs) :
        self.dirpath = dirpath.resolve()
        self.data_files_by_path = {}
        kwargs = populated_kwargs(kwargs,{'logger_file':self.dirpath})
        super().__init__(*args,**kwargs)



"""
A wrapper around KafkaCrypto provision scripts to automatically
move the output in the location expected by OpenMSIStream
"""

#imports
import pathlib, shutil, logging, warnings
from argparse import ArgumentParser
import urllib.request
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import kafkacrypto
from ..workflow.config import RUN_CONST
from .logging import Logger
from .config_file_parser import ConfigFileParser
from .misc import change_dir

#constants
LOGGER = Logger('ProvisionNode',logging.INFO)
KC_PATH = kafkacrypto.__path__
SP_NAME = 'simple-provision.py'
OP_NAME = 'online-provision.py'
GITHUB_URL = f'https://raw.githubusercontent.com/tmcqueen-materials/kafkacrypto/master/tools/{SP_NAME}'
TEMP_DIR_PATH = RUN_CONST.CONFIG_FILE_DIR/'temp_kafkacrypto_dir'

def get_script_location_and_code_from_input(args) :
    """
    Use a given path to get the script code
    """
    p_code = None
    p_loc = None
    if args.script_path.is_file() :
        p_loc = args.script_path
        with open(args.script_path,'rb') as fp :
            p_code = fp.read()
    #directory containing the script was given
    elif args.script_path.is_dir() :
        if (args.script_path/SP_NAME).is_file() :
            p_loc = args.script_path/SP_NAME
            with open(args.script_path/SP_NAME,'rb') as fp :
                p_code = fp.read()
        elif (args.script_path/OP_NAME).is_file() :
            p_loc = args.script_path/OP_NAME
            with open(args.script_path/OP_NAME,'rb') as fp :
                p_code = fp.read()
    return p_loc, p_code

def get_script_location_and_code(args) :
    """
    Return the location of the provision script and its text contents, given the arguments
    """
    p_code = None
    p_loc = None
    if args.script_path is not None:
        #path to the script itself was given
        p_loc, p_code = get_script_location_and_code_from_input(args)
    #if not set yet, try getting from the kafkacrypto install location (works if installed with --editable)
    if len(KC_PATH)==1 :
        if args.mode=='simple' :
            to_try = pathlib.Path(KC_PATH[0]).parent/'tools'/SP_NAME
        elif args.mode=='online' :
            to_try = pathlib.Path(KC_PATH[0]).parent/'tools'/OP_NAME
        if to_try.is_file() :
            p_loc = to_try
            with open(to_try,'rb') as fp :
                p_code = fp.read()
    if p_code is None :
        #if all else fails, try getting from the Github webpage (only simple-provision can be fetched this way)
        if args.mode=='simple' :
            try :
                with urllib.request.urlopen(GITHUB_URL) as urlp :
                    p_code = urlp.read()
                p_loc = GITHUB_URL
            except Exception :
                pass
    if p_loc is None or p_code is None :
        LOGGER.error('ERROR: failed to find the provisioning script to use!',RuntimeError)
    return p_loc, p_code

def move_files(p_loc) :
    """
    Move the created files into the expected locations
    """
    try :
        new_files = {}
        exts = ['.config','.seed','.crypto']
        for ext in exts :
            n_files = 0
            for fp in TEMP_DIR_PATH.glob(f'*{ext}') :
                n_files+=1
                new_files[ext] = fp.resolve()
            if n_files!=1 :
                LOGGER.error(f'ERROR: found {n_files} new {ext} files in {TEMP_DIR_PATH}',RuntimeError)
        node_id = None
        for ext in exts :
            filename = new_files[ext].name
            this_node_id = ((filename).split(ext))[0]
            if node_id is None :
                node_id = this_node_id
            elif node_id!=this_node_id :
                errmsg = f'ERROR: found a file called {filename} that conflicts with node_id {node_id}!'
                LOGGER.error(errmsg,RuntimeError)
        cfp = ConfigFileParser(new_files['.config'],logger=LOGGER)
        default_dict = cfp.get_config_dict_for_groups('DEFAULT')
        if 'node_id' not in default_dict :
            LOGGER.error(f"ERROR: node_id not listed in {new_files['.config']}!")
        elif default_dict['node_id']!=node_id :
            LOGGER.error(f"ERROR: node_id listed in {new_files['.config']} mismatched to filenames ({node_id})!")
        new_dirpath = TEMP_DIR_PATH.parent/node_id
        TEMP_DIR_PATH.rename(new_dirpath)
        LOGGER.info(f'Successfuly set up new KafkaCrypto node called "{node_id}"')
    except Exception as exc :
        errmsg = f'ERROR: Running {p_loc} did not produce the expected output! Temporary directories '
        errmsg+= f'will be removed and you will need to try again. Exception: {exc}'
        LOGGER.error(errmsg)
    finally :
        if TEMP_DIR_PATH.is_dir() :
            shutil.rmtree(TEMP_DIR_PATH)

def main() :
    """
    Script to run either simple or online provision for KafkaCrypto and
    move the output to the location expected by OpenMSIStream
    """
    #command line arguments
    parser = ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--mode', choices=['simple','online'], default='simple',
                       help='''Choice of which known type of provisioning to use.
                               Can also provide a path to the script to run instead.''')
    group.add_argument('--script-path', type=pathlib.Path, default='.',
                        help='Path to the provision script to run')
    args = parser.parse_args()
    if TEMP_DIR_PATH.is_dir() :
        shutil.rmtree(TEMP_DIR_PATH)
    #get the location/content of the simple-provision script
    p_loc, p_code = get_script_location_and_code(args)
    #run the script
    try :
        if not TEMP_DIR_PATH.is_dir() :
            TEMP_DIR_PATH.mkdir(parents=True)
        with change_dir(TEMP_DIR_PATH) :
            exec(p_code)
    except Exception as exc :
        LOGGER.error(f'ERROR: failed to run provisioning using {p_loc}! Exception: {exc}',RuntimeError)
    #make sure required files exist and move them into a new directory named for the node_ID
    move_files(p_loc)

if __name__=='__main__' :
    main()



"""A ConsumerGroup whose receipt of messages is governed using the ControlledProcess infrastructure"""

#imports
import time
from abc import ABC, abstractmethod
from ..workflow.controlled_process_multi_threaded import ControlledProcessMultiThreaded
from .consumer_group import ConsumerGroup

class ControlledMessageProcessor(ControlledProcessMultiThreaded,ConsumerGroup,ABC) :
    """
    Combine a ControlledProcessMultiThreaded and a ConsumerGroup to create a
    single interface for reading and processing individual messages
    """

    CONSUMER_POLL_TIMEOUT = 0.050
    NO_MESSAGE_WAIT = 0.005 #how long to wait if consumer.get_next_message_value returns None

    def __init__(self,*args,**kwargs) :
        """
        Hang onto the number of messages read and processed
        """
        self.n_msgs_read = 0
        self.n_msgs_processed = 0
        super().__init__(*args,**kwargs)
        self.restart_at_beginning = False #set to true to reset new consumers to their earliest offsets
        self.message_key_regex = None #set to some regex to filter messages by their keys
        self.filter_new_messages = False #reset the regex after the consumer has filtered through previous messages
        self.last_message = None #hold onto the last consumed message to manually commit its offset on shutdown

    def _run_worker(self):
        """
        Handle startup and shutdown of a thread-independent Consumer and
        serve individual messages to the _process_message function
        """
        #create the Consumer for this thread
        if self.alive :
            consumer = self.get_new_subscribed_consumer(restart_at_beginning=self.restart_at_beginning,
                                                        message_key_regex=self.message_key_regex,
                                                        filter_new_messages=self.filter_new_messages)
        if ('enable.auto.commit' not in consumer.configs.keys()) or (consumer.configs['enable.auto.commit'] is True) :
            warnmsg = 'WARNING: enable.auto.commit has not been set to False for a Consumer that will manually commit '
            warnmsg+= 'offsets. Missed or duplicate messages could result. You can set "enable.auto.commit"=False in '
            warnmsg+= 'the "consumer" section of the config file to re-enable manual offset commits (recommended).'
            self.logger.warning(warnmsg)
        #start the loop for while the controlled process is alive
        while self.alive :
            self.__do_alive_loop_iteration(consumer)
        #commit the offset of the last message received
        if (self.last_message is not None) and (not consumer.message_consumed_before(self.last_message)) :
            self.__commit_last_message_offset(consumer)
        #shut down the Consumer that was created once the process isn't alive anymore
        consumer.close()

    def _on_shutdown(self):
        super()._on_shutdown()
        self.close()

    @abstractmethod
    def _process_message(self,lock,msg,*args,**kwargs) :
        """
        Process a single message read from the thread-independent Consumer
        Returns true if processing was successful, and False otherwise

        lock = lock across all created child threads (use to enforce thread safety during processing)
        msg  = a single message that was consumed and should be processed by this function

        Not implemented in the base class
        """
        raise NotImplementedError

    def __do_alive_loop_iteration(self,consumer) :
        """
        Helper function to perform actions that happen continuously while the process is alive
        """
        #consume a message from the topic
        msg = consumer.get_next_message(self.CONSUMER_POLL_TIMEOUT)
        if msg is None :
            time.sleep(self.NO_MESSAGE_WAIT) #wait just a bit to not over-tax things
            return
        with self.lock :
            self.n_msgs_read+=1
            self.last_message = msg
        #send the message to the _process_message function
        retval = self._process_message(self.lock,msg)
        #count and (asynchronously) commit the message as processed (if it wasn't consumed already in the past)
        if retval :
            with self.lock :
                self.n_msgs_processed+=1
            if not consumer.message_consumed_before(msg) :
                tps = consumer.commit(msg)
                if tps is None :
                    return
                for t_p in tps :
                    if t_p.error is not None :
                        warnmsg = 'WARNING: failed to synchronously commit offset of last message received on '
                        warnmsg+= f'"{t_p.topic}" partition {t_p.partition}. Duplicate messages may result '
                        warnmsg+= f'when this Consumer is restarted. Error reason: {t_p.error.str()}'
                        self.logger.warning(warnmsg)

    def __commit_last_message_offset(self,consumer) :
        """
        Commit the offset of the last message received if it wasn't already consumed in the past (block until done)
        """
        try :
            tps = consumer.commit(self.last_message,asynchronous=False)
            if tps is not None :
                for t_p in tps :
                    if t_p.error is not None :
                        warnmsg = 'WARNING: failed to synchronously commit offset of last message received on '
                        warnmsg+= f'"{t_p.topic}" partition {t_p.partition}. Duplicate messages may result '
                        warnmsg+= f'when this Consumer is restarted. Error reason: {t_p.error.str()}'
                        self.logger.warning(warnmsg)
        except Exception as exc :
            errmsg = 'WARNING: failed to synchronously commit offset of last message received. '
            errmsg+= 'Duplicate messages may be read the next time this Consumer is started. '
            errmsg+= 'Error will be logged below but not re-raised.'
            self.logger.error(errmsg,exc_obj=exc,reraise=False)



"""Transfer contents of DataFiles read from chunks in a topic to an S3 bucket when complete files become available"""

#imports
from ..data_file_io.actor.data_file_stream_processor import DataFileStreamProcessor
from .config_file_parser import S3ConfigFileParser
from .s3_data_transfer import S3DataTransfer

class S3TransferStreamProcessor(DataFileStreamProcessor) :
    """
    A class to reconstruct data files read as messages from a topic, hold them in memory,
    and transfer them to an S3 bucket when all of their messages have been received

    :param bucket_name: Name of the S3 bucket into which reconstructed files should be transferred
    :type bucket_name: str
    :param config_path: Path to the config file to use in defining the Broker connection and Consumers
    :type config_path: :class:`pathlib.Path`
    :param topic_name: Name of the topic to which the Consumers should be subscribed
    :type topic_name: str
    """

    def __init__(self, bucket_name, config_path, topic_name, **kwargs) :
        super().__init__(config_path, topic_name, **kwargs)
        parser = S3ConfigFileParser(config_path,logger=self.logger)
        self.__s3_config = parser.s3_configs
        self.__s3_config['bucket_name'] = bucket_name
        self.bucket_name = bucket_name
        self.s3d = S3DataTransfer(self.__s3_config,logger=self.logger)

    def make_stream(self):
        """
        Runs :func:`~DataFileStreamProcessor.process_files_as_read` to reconstruct files in memory
        and transfer completed files to the S3 bucket. Runs until the user inputs a command to shut it down.

        :return: the total number of messages consumed
        :rtype: int
        :return: the total number of messages processed (registered in memory)
        :rtype: int
        :return: the paths of files successfully transferred to the S3 bucket during the run
        :rtype: list
        """
        return self.process_files_as_read()

    def _process_downloaded_data_file(self, datafile, lock):
        """
        Transfer a fully-reconstructed file to the S3 bucket and verify that its contents in the bucket
        match its original hash from disk. Logs a warning if the file hashes don't match.

        :param datafile: A :class:`~DownloadDataFileToMemory` object that has received
            all of its messages from the topic
        :type datafile: :class:`~DownloadDataFileToMemory`
        :param lock: Acquiring this :class:`threading.Lock` object would ensure that only one instance
            of :func:`~_process_downloaded_data_file` is running at once
        :type lock: :class:`threading.Lock`

        :return: None if processing was successful, a caught Exception otherwise
        """
        object_key = self.__get_datafile_object_key(datafile)
        try :
            self.s3d.transfer_object_stream(object_key, datafile)
        except Exception as exc :
            self.logger.error(f'ERROR: failed to transfer {datafile.filename} to the object store')
            return exc
        if self.s3d.compare_consumer_datafile_with_s3_object_stream(self.bucket_name, object_key, datafile):
            self.logger.info(object_key + ' matched with consumer datafile')
            # self.s3d.delete_object_from_bucket(self.bucket_name, object_key)
        else :
            warnmsg = f'WARNING: {object_key} transferred to bucket but the file on the bucket does not match '
            warnmsg+= 'the file originally read from disk!'
            self.logger.warning(warnmsg)
        return None

    def __get_datafile_object_key(self,datafile) :
        file_name = str(datafile.filename)
        sub_dir = datafile.subdir_str
        object_key = self.topic_name
        if sub_dir!='' :
            object_key+= '/' + sub_dir
        object_key+= '/' + file_name
        return object_key

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = [*superargs, 'bucket_name']
        kwargs = {**superkwargs,'config':'test_s3_transfer'}
        return args, kwargs

    @classmethod
    def run_from_command_line(cls, args=None):
        """
        Run a :class:`~S3TransferStreamProcessor` directly from the command line

        Calls :func:`~make_stream` on a :class:`~S3TransferStreamProcessor` defined by
        command line (or given) arguments

        :param args: the list of arguments to send to the parser instead of getting them from sys.argv
        :type args: list, optional
        """
        # make the argument parser
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        s3_stream_proc = cls(args.bucket_name,
                             args.config, args.topic_name,
                             output_dir=args.output_dir,
                             n_threads=args.n_threads,
                             update_secs=args.update_seconds,
                             consumer_group_id=args.consumer_group_id)
        # cls.bucket_name = args.bucket_name
        msg = f'Listening to the {args.topic_name} topic for files to add to the {args.bucket_name} bucket....'
        s3_stream_proc.logger.info(msg)
        n_read,n_processed,complete_filenames = s3_stream_proc.make_stream()
        s3_stream_proc.close()
        msg = f'{n_read} total messages were consumed'
        if len(complete_filenames)>0 :
            msg+=f', {n_processed} messages were successfully processed,'
            msg+=f' and the following {len(complete_filenames)} file'
            msg+=' was' if len(complete_filenames)==1 else 's were'
            msg+=f' successfully transferred to the {args.bucket_name} bucket'
            for fn in complete_filenames :
                msg+=f'\n\t{fn}'
        else :
            msg+=f' and {n_processed} messages were successfully processed'
        s3_stream_proc.logger.info(msg)

    def _on_check(self) :
        msg = f'{self.n_msgs_read} messages read, {self.n_msgs_processed} messages processed, '
        msg+= f'{len(self.completely_processed_filepaths)} files transferred so far'
        self.logger.debug(msg)
        if len(self.files_in_progress_by_path)>0 or len(self.completely_processed_filepaths)>0 :
            self.logger.debug(self.progress_msg)

def main(args=None):
    """
    Main method to run from command line
    """
    S3TransferStreamProcessor.run_from_command_line(args)

if __name__ == '__main__':
    main()



"""Various types of DataFiles that have been read back from messages in a topic"""

#imports
import os
from hashlib import sha512
from contextlib import nullcontext
from abc import ABC, abstractmethod
from ..config import DATA_FILE_HANDLING_CONST
from .data_file import DataFile

class DownloadDataFile(DataFile,ABC) :
    """
    Class to represent a data file that will be read as messages from a topic

    :param filepath: Path to the file
    :type filepath: :class:`pathlib.Path`
    """

    #################### PROPERTIES AND STATIC METHODS ####################

    @staticmethod
    def get_full_filepath(dfc) :
        """
        Return the full filepath of a downloaded file given one of its DataFileChunks

        :param dfc: One of the DataFileChunk objects contributing to the file
        :type dfc: :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk`

        :return: the full path to the file
        :rtype: :class:`pathlib.Path`
        """
        if dfc.filename_append=='' :
            return dfc.filepath
        filename_split = dfc.filepath.name.split('.')
        full_fp = dfc.filepath.parent/(filename_split[0]+dfc.filename_append+'.'+('.'.join(filename_split[1:])))
        return full_fp

    @property
    @abstractmethod
    def check_file_hash(self) :
        """
        The hash of the data in the file after it was read. Not implemented in the base class.
        """
        raise NotImplementedError

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,filepath,*args,**kwargs) :
        super().__init__(filepath,*args,**kwargs)
        #start an empty set of this file's downloaded offsets
        self._chunk_offsets_downloaded = []
        self.full_filepath = None
        self.subdir_str = None
        self.n_total_chunks = None

    def add_chunk(self,dfc,thread_lock=nullcontext()) :
        """
        Process a chunk that's been read from a topic.
        Returns a number of codes based on what effect adding the chunk had.

        This function calls :func:`~_on_add_chunk`,
        with the :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk` as the argument.

        :param dfc: the DataFileChunk object whose data should be added
        :type dfc: :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk`
        :param thread_lock: the lock object to acquire/release so that race conditions don't affect reconstruction
                            of the files (only needed if running this function asynchronously)
        :type thread_lock: :class:`threading.Lock`, optional

        :return: an internal code indicating whether the chunk: was successfully added to a file in progress,
            was already received, was the last chunk needed and the file is successfully reconstructed according
            to its hash or the post-reconstruction has is mismatched to the hash of the file contents originally
            read from disk.
        :rtype: int
        """
        #if this chunk's offset has already been written to disk, return the "already written" code
        with thread_lock :
            already_written = dfc.chunk_offset_write in self._chunk_offsets_downloaded
        if already_written :
            return DATA_FILE_HANDLING_CONST.CHUNK_ALREADY_WRITTEN_CODE
        #the filepath of this DownloadDataFile and of the given DataFileChunk must match
        if dfc.filepath!=self.filepath :
            errmsg = f'ERROR: filepath mismatch between data file chunk with {dfc.filepath} and '
            errmsg+= f'data file with {self.filepath}'
            self.logger.error(errmsg,ValueError)
        #modify the filepath to include any append to the name
        full_filepath = self.__class__.get_full_filepath(dfc)
        if self.full_filepath is None :
            self.full_filepath = full_filepath
            self.filename = self.full_filepath.name
        elif self.full_filepath!=full_filepath :
            errmsg = f'ERROR: filepath for data file chunk {dfc.chunk_i}/{dfc.n_total_chunks} with offset '
            errmsg+= f'{dfc.chunk_offset_write} is {full_filepath} but the file being reconstructed is '
            errmsg+= f'expected to have filepath {self.full_filepath}'
            self.logger.error(errmsg,ValueError)
        #add the subdirectory string to this file
        if self.subdir_str is None :
            self.subdir_str = dfc.subdir_str
        elif self.subdir_str!=dfc.subdir_str :
            errmsg = f"Mismatched subdirectory strings! From file = {self.subdir_str}, from chunk = {dfc.subdir_str}"
            self.logger.error(errmsg,ValueError)
        #set or check the total number of chunks expected
        if self.n_total_chunks is None :
            with thread_lock :
                self.n_total_chunks = dfc.n_total_chunks
        elif self.n_total_chunks!=dfc.n_total_chunks :
            errmsg = f'ERROR: {self.__class__.__name__} with filepath {self.full_filepath} is expecting '
            errmsg+= f'{self.n_total_chunks} chunks but found a chunk from a split with '
            errmsg+= f'{dfc.n_total_chunks} total chunks.'
            self.logger.error(errmsg,ValueError)
        #acquire the thread lock to make sure this process is the only one dealing with this particular file
        with thread_lock :
            #call the function to actually add the chunk
            self._on_add_chunk(dfc)
            #add the offset of the added chunk to the set of reconstructed file chunks
            self._chunk_offsets_downloaded.append(dfc.chunk_offset_write)
            last_chunk = len(self._chunk_offsets_downloaded)==dfc.n_total_chunks
        #if this chunk was the last that needed to be added, check the hashes
        if last_chunk :
            if self.check_file_hash!=dfc.file_hash :
                return DATA_FILE_HANDLING_CONST.FILE_HASH_MISMATCH_CODE
            return DATA_FILE_HANDLING_CONST.FILE_SUCCESSFULLY_RECONSTRUCTED_CODE
        return DATA_FILE_HANDLING_CONST.FILE_IN_PROGRESS

    #################### PRIVATE HELPER FUNCTIONS ####################

    @abstractmethod
    def _on_add_chunk(self,dfc) :
        """
        A function to actually process a new chunk being added to the file.
        This function is executed while a thread lock is acquired so it will never run asynchronously.
        Also any DataFileChunks passed to this function are guaranteed to have unique offsets.

        Not implemented in the base class.

        :param dfc: the DataFileChunk object whose data should be added
        :type dfc: :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk`
        """
        raise NotImplementedError

class DownloadDataFileToDisk(DownloadDataFile) :
    """
    Class to represent a data file that will be reconstructed on disk using messages read from a topic

    :param filepath: Path to the file
    :type filepath: :class:`pathlib.Path`
    """

    #################### PROPERTIES ####################

    @property
    def check_file_hash(self) :
        """
        Hash of the file contents as read from its current location on disk
        """
        check_file_hash = sha512()
        with open(self.full_filepath,'rb') as fp :
            data = fp.read()
        check_file_hash.update(data)
        return check_file_hash.digest()

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,filepath,*args,**kwargs) :
        super().__init__(filepath,*args,**kwargs)
        #create the parent directory of the file if it doesn't exist yet (in case the file is in a new subdirectory)
        if not self.filepath.parent.is_dir() :
            self.filepath.parent.mkdir(parents=True)

    def _on_add_chunk(self,dfc) :
        """
        Add the data from a given file chunk to this file on disk

        :param dfc: the DataFileChunk object whose data should be added
        :type dfc: :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk`
        """
        mode = 'r+b' if self.full_filepath.is_file() else 'w+b'
        with open(self.full_filepath,mode) as fp :
            fp.seek(dfc.chunk_offset_write)
            fp.write(dfc.data)
            fp.flush()
            os.fsync(fp.fileno())
            fp.close()

class DownloadDataFileToMemory(DownloadDataFile) :
    """
    Class to represent a data file that will be held in memory and populated by the contents of messages from a topic

    :param filepath: Path to the file
    :type filepath: :class:`pathlib.Path`
    """

    #################### PROPERTIES ####################

    @property
    def bytestring(self) :
        """
        The bytestring of the file contents (like calling file_pointer.read() for a file opened in "rb" mode).
        Call bytestring.decode() to convert this to a text string.
        """
        if self.__bytestring is None :
            self.__create_bytestring()
        return self.__bytestring

    @property
    def check_file_hash(self) :
        """
        The hash of the file contents determined from the bytestring of the file stored in memory
        """
        check_file_hash = sha512()
        check_file_hash.update(self.bytestring)
        return check_file_hash.digest()

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,filepath,*args,**kwargs) :
        super().__init__(filepath,*args,**kwargs)
        #start a dictionary of the file data by their offsets
        self.__chunk_data_by_offset = {}
        #placeholder for the eventual full data bytestring
        self.__bytestring = None

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _on_add_chunk(self,dfc) :
        """
        Add the data from a given file chunk to the dictionary of data by offset

        :param dfc: the DataFileChunk object whose data should be added
        :type dfc: :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk`
        """
        self.__chunk_data_by_offset[dfc.chunk_offset_write] = dfc.data

    def __create_bytestring(self) :
        """
        Makes all of the data held in the dictionary into a single bytestring ordered by offset of each chunk
        """
        bytestring = b''
        for data in [self.__chunk_data_by_offset[offset] for offset in sorted(self.__chunk_data_by_offset.keys())] :
            bytestring+=data
        self.__bytestring = bytestring



"""A single chunk of a DataFile. Can be produced to a topic or consumed from one."""

#imports
import pathlib
from hashlib import sha512
from ...utilities.logging import Logger
from ...kafka_wrapper.producible import Producible
from ..utilities import get_message_prepend

class DataFileChunk(Producible) :
    """
    Class representing a single chunk of an uploaded or downloaded file. DataFileChunk objects are
    automatically serialized/deserialized when they are produced/consumed to topics using OpenMSIStream.

    :param filepath: path to this chunk's file (fully resolved if being produced, may be relative if it was consumed)
    :type filepath: :class:`pathlib.Path`
    :param filename: the name of the file
    :type filename: str
    :param file_hash: hash of this chunk's entire file data
    :type file_hash: str
    :param chunk_hash: hash of this chunk's data
    :type chunk_hash: str
    :param chunk_offset_read: offset (in bytes) of this chunk within the original file
    :type chunk_offset_read: int
    :param chunk_offset_write: offset (in bytes) of this chunk within the reconstructed file (may be different than
        chunk_offset_read due to excluding some bytes in uploading)
    :type chunk_offset_write: int
    :param chunk_size: size of this chunk (in bytes)
    :type chunk_size: int
    :param chunk_i: index of this chunk within the larger file
    :type chunk_i: int
    :param n_total_chunks: the total number of chunks to expect from the original file
    :type n_total_chunks: int
    :param rootdir: path to the "root" directory; anything beyond in the filepath is considered a subdirectory
        (can also be set later)
    :type rootdir: :class:`pathlib.Path`, optional
    :param filename_append: string to append to the stem of the filename when the file is reconstructed
    :type filename_append: str, optional
    :param data: the actual binary data of this chunk of the file (can be set later if this chunk is being produced
        and not consumed)
    :type data: bytes, optional
    """

    #################### PROPERTIES ####################

    @property
    def filepath(self) :
        """
        The path to the file
        """
        return self.__filepath

    @property
    def rootdir(self) :
        """
        The path to the file's root directory (already set if chunk is to be produced,
        but must be set later if chunk is a consumed message)
        """
        return self.__rootdir

    @rootdir.setter
    def rootdir(self,rootdir) :
        """
        Also resets the overall filepath (used in consuming messages for files in subdirectories)
        """
        self.__rootdir=rootdir
        if rootdir is not None :
            try :
                self.__filepath = self.__filepath.relative_to(self.__rootdir)
            except ValueError :
                pass
            self.__filepath = self.__rootdir / self.__filepath

    @property
    def subdir_str(self) :
        """
        A string representation of the path to the file, relative to its root directory
        """
        if self.__rootdir is None :
            parentdir_as_posix = self.__filepath.parent.as_posix()
            if parentdir_as_posix=='.' :
                return ''
            return parentdir_as_posix
        relpath = self.__filepath.parent.relative_to(self.__rootdir)
        if relpath==pathlib.Path() :
            return ''
        return relpath.as_posix()

    @property
    def msg_key(self) :
        """
        string representing the key of the message this chunk will be produced as
        """
        key_pp = get_message_prepend(self.subdir_str,self.filename)
        return f'{key_pp}_{self.chunk_i}_of_{self.n_total_chunks}'

    @property
    def msg_value(self) :
        """
        value of the message this chunk will be produced as (just the object itself, since a
        :class:`DataFileChunkSerializer` is used)
        """
        return self

    @property
    def callback_kwargs(self):
        """
        keyword arguments that should be sent to the producer callback function when the chunk is produced
        """
        return {
            'filepath' : self.__filepath,
            'filename' : self.filename,
            'n_total_chunks' : self.n_total_chunks,
            'chunk_i' : self.chunk_i,
            }


    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,filepath,filename,file_hash,chunk_hash,chunk_offset_read,chunk_offset_write,chunk_size,chunk_i,
                 n_total_chunks,rootdir=None,filename_append='',data=None) :
        self.__filepath = filepath
        self.filename = filename
        self.file_hash = file_hash
        self.chunk_hash = chunk_hash
        self.chunk_offset_read = chunk_offset_read
        self.chunk_offset_write = chunk_offset_write
        self.chunk_size = chunk_size
        self.chunk_i = chunk_i
        self.n_total_chunks = n_total_chunks
        self.__rootdir = rootdir
        self.filename_append = filename_append
        self.data = data

    def __eq__(self,other) :
        if not isinstance(other,DataFileChunk) :
            return NotImplemented
        #compare everything but the filepath
        retval = self.filename == other.filename
        retval = retval and self.file_hash == other.file_hash
        retval = retval and self.chunk_hash == other.chunk_hash
        retval = retval and self.chunk_offset_write == other.chunk_offset_write
        retval = retval and self.chunk_size == other.chunk_size
        retval = retval and self.chunk_i == other.chunk_i
        retval = retval and self.n_total_chunks == other.n_total_chunks
        retval = retval and self.subdir_str == other.subdir_str
        retval = retval and self.filename_append == other.filename_append
        retval = retval and self.data == other.data
        return retval

    def __str__(self) :
        string_rep = 'DataFileChunk('
        string_rep+=f'filename: {self.filename}, '
        string_rep+=f'file_hash: {self.file_hash}, '
        string_rep+=f'chunk_hash: {self.chunk_hash}, '
        string_rep+=f'chunk_offset_read: {self.chunk_offset_read}, '
        string_rep+=f'chunk_offset_write: {self.chunk_offset_write}, '
        string_rep+=f'chunk_size: {self.chunk_size}, '
        string_rep+=f'chunk_i: {self.chunk_i}, '
        string_rep+=f'n_total_chunks: {self.n_total_chunks}, '
        string_rep+=f'subdir_str: {self.subdir_str}, '
        string_rep+=f'filename_append: {self.filename_append}, '
        #string_rep+=f'data: {self.data}, '
        string_rep+=')'
        return string_rep

    def __hash__(self) :
        return super().__hash__()

    def get_log_msg(self, print_every=None):
        """
        If the chunk's index mod ``print_every`` is 0, or if the chunk is the last one for the file,
        returns a message to log. Otherwise returns None.

        :param print_every: number of chunks that should be skipped between logging messages
        :type print_every: int, optional

        :return: message to log stating which chunk from which file is being uploaded
        :rtype: str, or None
        """
        if (print_every and (self.chunk_i-1)%print_every==0) or self.chunk_i==self.n_total_chunks :
            return f'uploading {self.filename} chunk {self.chunk_i} (out of {self.n_total_chunks})'
        return None

    def populate_with_file_data(self,logger=None) :
        """
        Populate this chunk with the actual data from the file. Called only when this chunk is being produced.

        :param logger: a logger object to use to log errors that may arise in populating the file chunk
        :type logger: :class:`~.utilities.Logger`, optional

        :raises FileNotFoundError: if the file doesn't exist on disk at ``self.filepath``
        :raises ValueError: if the data read from the file is not of the expected size,
            or if its hash is not matched to what was originally calculated.
        """
        #create a new logger if one isn't given
        if logger is None :
            logger = Logger(self.__class__.__name__)
        #make sure the file exists
        if not self.filepath.is_file() :
            logger.error(f'ERROR: file {self.filepath} does not exist!',FileNotFoundError)
        #get the data from the file
        with open(self.filepath, "rb") as fp:
            fp.seek(self.chunk_offset_read)
            data = fp.read(self.chunk_size)
        #make sure it's of the expected size
        if len(data) != self.chunk_size:
            msg = f'ERROR: chunk {self.chunk_hash} size {len(data)} != expected size {self.chunk_size} in file '
            msg+= f'{self.filepath}, offset {self.chunk_offset_read}'
            logger.error(msg,ValueError)
        #check that its hash matches what was found at the time of putting it in the queue
        check_chunk_hash = sha512()
        check_chunk_hash.update(data)
        check_chunk_hash = check_chunk_hash.digest()
        if self.chunk_hash != check_chunk_hash:
            msg = f'ERROR: chunk hash {check_chunk_hash} != expected hash {self.chunk_hash} in file {self.filepath}, '
            msg+= f'offset {self.chunk_offset_read}'
            logger.error(msg,ValueError)
        #set the chunk's data value
        self.data = data



"""Utilities for working with S3 buckets"""

# @author: Amir H. Sharifzadeh, https://pages.jh.edu/asharif9/
# @project: OpenMSI
# @date: 04/12/2022
# @owner: The Institute of Data Intensive Engineering and Science, https://idies.jhu.edu/
# Johns Hopkins University http://www.jhu.edu/
import hashlib
import boto3
from ..utilities import LogOwner

class S3Service(LogOwner) :
    """
    A class to work with an S3 connection session
    """

    def __init__(self, s3_config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = boto3.session.Session()

        self.bucket_name = s3_config['bucket_name']
        endpoint_url = str(s3_config['endpoint_url'])

        if not endpoint_url.startswith('https://'):
            endpoint_url = 'https://' + endpoint_url
        self.s3_client = self.session.client(
            service_name='s3',
            aws_access_key_id=s3_config['access_key_id'],
            aws_secret_access_key=s3_config['secret_key_id'],
            region_name=s3_config['region'],
            endpoint_url= endpoint_url
        )
        self.region = s3_config['region']
        self.grant_read = 'uri="http://acs.amazonaws.com/groups/global/AllUsers"'

    def get_object_stream_by_object_key(self, bucket_name, object_key):
        """
        Return the stream of an object in a bucket given its key
        """
        s3_response_object = self.s3_client.get_object(Bucket=bucket_name, Key=object_key)
        return s3_response_object['Body'].read()

    def delete_object_from_bucket(self, bucket_name, object_key):
        """
        Delete an object with a given key from a bucket
        """
        try:
            # make sure object exits before deleting the object from the bucket
            s3_response_object = self.s3_client.get_object(Bucket=bucket_name, Key=object_key)
            if s3_response_object is None:
                msg = 'The object ' + object_key + ' does not exist in this bucket: ' + bucket_name
                self.logger.error(msg)
                return
        except Exception:
            msg  = 'A problem occurred while reading ' + object_key + ' from bucket: ' + bucket_name
            self.logger.error(msg)
            return
        # delete the object from the bucket safely
        try:
            self.s3_client.delete_object(Bucket=bucket_name, Key=object_key)
            msg = object_key + ' was deleted successfully from bucket: ' + bucket_name
            self.logger.info(msg)
        except Exception:
            msg = 'Could not delete ' + object_key + ' from bucket: ' + bucket_name
            self.logger.error(msg)
            return

    def compare_producer_datafile_with_s3_object_stream(self, bucket_name, object_key, hashed_datafile_stream):
        """
        Compare a given DataFile hash with the hash of an object in a bucket
        """
        if hashed_datafile_stream is None:
            return False
        s3_response_object = self.s3_client.get_object(Bucket=bucket_name, Key=object_key)
        object_content = s3_response_object['Body'].read()
        md5 = hashlib.md5()
        md5.update(object_content)
        if object_content is None:
            return False
        hhh = format(md5.hexdigest())
        return str(hhh) == hashed_datafile_stream

    def compare_consumer_datafile_with_s3_object_stream(self, bucket_name, object_key, datafile):
        """
        Compare a given DataFile with an object in a bucket using their hashes
        """
        if datafile is None:
            return False
        s3_response_object = self.s3_client.get_object(Bucket=bucket_name, Key=object_key)
        object_content = s3_response_object['Body'].read()
        if object_content is None:
            return False
        datafile_md5 = hashlib.md5()
        datafile_md5.update(datafile.bytestring)
        object_content_md5 = hashlib.md5()
        object_content_md5.update(object_content)
        return format(datafile_md5.hexdigest()) == format(object_content_md5.hexdigest())



"""Constants used in running different controlled processes in general"""

#imports
import pathlib

class RunConstants :
    """
    Constants used for running controlled processes, reading config files, etc.
    """

    CONFIG_FILE_EXT = '.config'
    CONFIG_FILE_DIR = pathlib.Path(__file__).parent.parent / 'kafka_wrapper' / 'config_files'
    DEFAULT_N_THREADS = 2
    DEFAULT_UPDATE_SECONDS = 300

RUN_CONST = RunConstants()



"""Functionality common to both Windows Services and Linux daemons"""

#imports
import sys, pathlib, textwrap, importlib
from abc import abstractmethod
from subprocess import CalledProcessError
from ..utilities import LogOwner
from ..utilities.config_file_parser import ConfigFileParser
from ..workflow import Runnable, OpenMSIStreamArgumentParser
from ..workflow.has_argument_parser import HasArgumentParser
from .config import SERVICE_CONST
from .utilities import set_env_vars, remove_env_var

class ServiceManagerBase(LogOwner,HasArgumentParser) :
    """
    Base class for working with Services in general

    :param service_name: The name of the Service/daemon as installed
    :type service_name: str
    :param service_spec_string: A string specifying which code should be run as a Service.
        Could be the name of an OpenMSIStream Runnable class, or the path to a custom Python code.
        Custom Services can also specify a :class:`~Runnable` class name, and/or a function in the file
        using special formatting like [class_name]=[path.to.file]:[function_name].
        Only needed to initially install the Service/daemon.
    :type service_spec_string: str, optional
    :param argslist: The list of arguments (as from the command line) to pass to the :class:`~Runnable` class.
        Only needed to initially install the Service/daemon.
    :type argslist: list, optional
    :param interactive: if True, a few more messages/prompts will come up telling a user what to do
    :type interactive: bool, optional
    """

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,service_name,*,service_spec_string=None,argslist=None,interactive=None,**kwargs) :
        super().__init__(**kwargs)
        self.service_name = service_name
        self.service_spec_string = service_spec_string
        self.__set_service_dict()
        self.argslist = argslist
        self.interactive = interactive
        self.env_var_filepath = SERVICE_CONST.WORKING_DIR/f'{self.service_name}_env_vars.txt'
        self.install_args_filepath = SERVICE_CONST.WORKING_DIR/f'{self.service_name}_install_args.txt'
        self.exec_fp = SERVICE_CONST.WORKING_DIR/f'{self.service_name}{SERVICE_CONST.SERVICE_EXECUTABLE_NAME_STEM}'
        self.env_vars_needed = None

    def install_service(self) :
        """
        Install the Service
        child classes should call this method to get the setup done before they do anything specific
        """
        #make sure the necessary information was supplied
        if self.service_dict is None or self.argslist is None :
            errmsg = 'ERROR: newly installing a Service requires that the specifying what to install '
            errmsg+= 'and giving an argslist!'
            self.logger.error(errmsg,RuntimeError)
        #set the environment variables
        must_rerun = set_env_vars(self.env_var_names,interactive=self.interactive)
        if must_rerun :
            msg = 'New values for environment variables have been set. '
            msg+= 'Please close this window and rerun InstallService so that their values get picked up.'
            self.logger.info(msg)
            sys.exit(0)
        #write out the environment variable file
        self.env_vars_needed = self._write_env_var_file()
        #write out the install arguments file
        self._write_install_args_file()
        #write out the executable file
        self._write_executable_file()
        #install the service
        msg = 'Installing '
        if self.service_dict['class_name'] is not None :
            msg+='a'
            if self.service_dict['class_name'].lower() in ('a','e','i','o','u','y') :
                msg+='n'
            msg+=f" {self.service_dict['class_name']} program "
        else :
            msg+=f"{self.service_dict['filepath']}"
            if self.service_dict['func_name'] is not None :
                msg+=f":{self.service_dict['func_name']}"
        msg+=f' as "{self.service_name}" from executable at {self.exec_fp}'
        self.logger.info(msg)

    def run_manage_command(self,run_mode,remove_env_vars=False,remove_install_args=False,remove_nssm=False) :
        """
        Run one of the other functions according to the action given

        :param run_mode: The string identifying the action(s) that should be taken
        :type run_mode: str
        :param remove_env_vars: if True, any environment variables needed by the Service/daemon will be removed.
            Only used when removing the Service/daemon.
        :type remove_env_vars: bool, optional.
        :param remove_install_args: if True, the file listing the arguments used to install the Service/daemon
            (to make it easier to re-install) will be removed. Only used when removing the Service/daemon.
        :type remove_install_args: bool, optional
        :param remove_nssm: if True, the NSSM executable will be removed. Only used when removing the Service/daemon.
        :type remove_nssm: bool, optional
        """
        if run_mode in ['start'] :
            self.start_service()
        if run_mode in ['status'] :
            self.service_status()
        if run_mode in ['stop','stop_and_remove','stop_and_reinstall','stop_and_restart'] :
            self.stop_service()
        if run_mode in ['remove','stop_and_remove','stop_and_reinstall','stop_and_restart'] :
            self.remove_service(remove_env_vars=remove_env_vars,
                                remove_install_args=remove_install_args,
                                remove_nssm=remove_nssm)
        if run_mode in ['reinstall','stop_and_reinstall','stop_and_restart'] :
            self.reinstall_service()
        if run_mode in ['stop_and_restart'] :
            self.start_service()

    @abstractmethod
    def start_service(self) :
        """
        start the Service
        not implemented in the base class
        """
        raise NotImplementedError

    @abstractmethod
    def service_status(self) :
        """
        get the status of the service
        not implemented in the base class
        """
        raise NotImplementedError

    @abstractmethod
    def stop_service(self) :
        """
        stop the Service
        not implemented in the base class
        """
        raise NotImplementedError

    def remove_service(self,remove_env_vars=False,remove_install_args=False,remove_nssm=False) :
        """
        Remove the Service. Child classes should call this after doing whatever they need to do to
        optionally remove environment variables.

        :param remove_env_vars: if True, any environment variables needed by the Service/daemon will be removed.
        :type remove_env_vars: bool, optional.
        :param remove_install_args: if True, the file listing the arguments used to install the Service/daemon
            (to make it easier to re-install) will be removed.
        :type remove_install_args: bool, optional
        :param remove_nssm: if True, the NSSM executable will be removed. Only used on Windows.
        :type remove_nssm: bool, optional
        """
        #remove the executable file
        if self.exec_fp.is_file() :
            self.exec_fp.unlink()
        #remove the environment variables that were set when the service was installed
        if remove_env_vars :
            try :
                for env_var_name in self.env_var_names :
                    remove_env_var(env_var_name)
                    self.logger.info(f'{env_var_name} environment variable successfully removed')
            except CalledProcessError :
                warnmsg = 'WARNING: failed to remove environment variables. You should remove any username/password '
                warnmsg+= 'environment variables manually even though the service is uninstalled!'
                self.logger.info(warnmsg)
        else :
            self.logger.info('Environment variables will be retained')
        #remove the actual environment variable file
        if self.env_var_filepath.is_file() :
            self.env_var_filepath.unlink()
        #remove the installation arguments file
        if remove_install_args :
            if self.install_args_filepath.is_file() :
                self.install_args_filepath.unlink()
                self.logger.info(f'Installation arguments file {self.install_args_filepath} deleted.')
        else :
            if self.install_args_filepath.is_file() :
                self.logger.info(f'Installation arguments file {self.install_args_filepath} will be retained')
        self.logger.info(f'Done removing {self.service_name}')
        _ = remove_nssm # appease pyflakes / pylint

    def reinstall_service(self) :
        """
        Install the service using the arguments defined in the service arguments file
        """
        if not self.install_args_filepath.is_file() :
            errmsg =  'ERROR: cannot reinstall service without an installation arguments file '
            errmsg+= f'at {self.install_args_filepath}!'
            self.logger.error(errmsg,FileNotFoundError)
        with open(self.install_args_filepath,'r') as fp :
            lines = fp.readlines()
        if not len(lines)>0 :
            errmsg = f'ERROR: installation arguments file {self.install_args_filepath} does not contain enough entries!'
            self.logger.error(errmsg,RuntimeError)
        self.service_spec_string = lines[0].strip()
        self.__set_service_dict()
        self.argslist = [line.strip() for line in lines[1:]] if len(lines)>1 else []
        if self.interactive :
            msg = 'Running this reinstall would be like running the following from the command line:\n'
            msg+= f'InstallService {self.service_spec_string} '
            for arg in self.argslist :
                msg+=f'{arg} '
            msg+=f'--service_name {self.service_name}\n'
            msg+='Does everything above look right? [(y)/n]: '
            check = input(msg)
            if check.lower().startswith('n') :
                msg = f'{self.service_name} will NOT be reinstalled. You can install it again by rerunning '
                msg+= 'InstallService on its own.'
                self.logger.info(msg)
                sys.exit(0)
            else :
                self.install_service()

    #################### PRIVATE HELPER FUNCTIONS ####################

    @abstractmethod
    def _write_env_var_file(self) :
        """
        Write and set permissions for the file holding the values of the environment variables needed by the Service
        returns True if there are environment variables referenced, False if not

        Not implemented in base class
        """
        raise NotImplementedError

    def _write_install_args_file(self) :
        """
        Write out a file storing the arguments used to install the Service so it can be
        reinstalled using the same configurations/setting if desired
        """
        if self.service_spec_string is None or self.argslist is None :
            errmsg = "ERROR: can't write the installation arguments file without a service_spec_string and argslist!"
            self.logger.error(errmsg,RuntimeError)
        with open(self.install_args_filepath,'w') as fp :
            for arg in [self.service_spec_string,*self.argslist] :
                fp.write(f'{arg}\n')

    def _write_executable_file(self,filepath=None) :
        """
        write out the executable python file that the service will actually be running
        """
        error_log_path = pathlib.Path().resolve()/f'{self.service_name}{SERVICE_CONST.ERROR_LOG_STEM}'
        code = '''\
            if __name__=='__main__' :
                try :'''
        if self.service_dict['func_name'] is not None :
            code+=f'''
                    from {self.service_dict['filepath']} import {self.service_dict['func_name']}
                    {self.service_dict['func_name']}({self.argslist})'''
        else :
            code+=f'''
                    from {self.service_dict['filepath']} import {self.service_dict['class_name']}
                    {self.service_dict['class_name']}.run_from_command_line({self.argslist})'''
        code+=f'''
                except Exception :
                    import pathlib, traceback, datetime
                    output_filepath = pathlib.Path(r"{error_log_path}")
                    with open(output_filepath,'a') as fp :'''
        code+=r'''
                        fp.write(f'{(datetime.datetime.now()).strftime("Error on %Y-%m-%d at %H:%M:%S")}. Exception:\n{traceback.format_exc()}')
                    import sys
                    sys.exit(1)
        '''
        if filepath is not None :
            warnmsg = f'WARNING: Services executable will be written to {filepath} instead of {self.exec_fp}'
            self.logger.warning(warnmsg)
            exec_fp = filepath
        else :
            exec_fp = self.exec_fp
        with open(exec_fp,'w') as fp :
            fp.write(textwrap.dedent(code))

    def __set_service_dict(self) :
        """
        Set the service dict with information about the code that should be run based on the service_spec_string
        """
        if self.service_spec_string is not None :
            service_dict = [sd for sd in SERVICE_CONST.available_services
                            if sd['class_name']==self.service_spec_string]
            if len(service_dict)==1 :
                self.service_dict = service_dict[0]
            elif len(service_dict)==0 :
                f_p, c_n, r_c, f_n = self.__parse_custom_service_string(self.service_spec_string,self.logger)
                self.service_dict = {'filepath':f_p,
                                     'class_name':c_n,
                                     'class':r_c,
                                     'func_name':f_n}
            else :
                errmsg = f'ERROR: could not find the Service dictionary for {self.service_name} '
                errmsg+= f'(a {self.service_spec_string} program)! service_dict = {service_dict}'
                self.logger.error(errmsg,RuntimeError)
        else :
            self.service_dict = None

    @staticmethod
    def __parse_custom_service_string(service_spec_string,logger=SERVICE_CONST.logger) :
        """
        Get the filepath and optional class/function names from the custom Service string
        """
        filepath = None
        class_name = None
        run_class = None
        func_name = None
        try :
            #at minimum need a path to a file containing a class or function to run
            if '=' in service_spec_string :
                equals_split = service_spec_string.split('=')
                assert len(equals_split)==2
                class_name = equals_split[0]
                for_path_and_func_name = equals_split[1]
            else :
                class_name = None
                for_path_and_func_name = service_spec_string
            if ':' in service_spec_string :
                colon_split = for_path_and_func_name.split(':')
                assert len(colon_split)==2
                filepath = colon_split[0]
                func_name = colon_split[1]
            else :
                filepath = for_path_and_func_name
            assert filepath is not None
            #make sure the path is valid
            module = importlib.import_module(filepath)
            assert module is not None
            #if the function name was specified, make sure that can be imported from the file, too
            if func_name is not None :
                function = getattr(module,func_name)
                assert function is not None
            #If the class name was given without a function name, make sure the class can be imported from the file
            elif class_name is not None and func_name is None :
                run_class = getattr(module,class_name)
                #and make sure the class extends Runnable, since we'll be calling its run_from_command_line function
                assert issubclass(run_class,Runnable)
            #make sure at least one of the function/class names was given
            assert ((func_name is not None) or ((class_name is not None) and (run_class is not None)))
        except Exception as exc :
            errmsg = f'ERROR: service specification string {service_spec_string} is not valid! '
            errmsg+= 'Will re-raise specific Exception.'
            logger.error(errmsg,exc_obj=exc)
        return filepath, class_name, run_class, func_name

    #################### CLASS METHODS ####################

    @classmethod
    def get_argument_parser(cls,install_or_manage=None,class_name_or_spec_string=None) :
        """
        Return the command line argument parser that should be used

        :param install_or_manage: Whether the parser used should accept commands for "install"-ing
            or "manage"-ing the Service/daemon
        :type install_or_manage: str
        :param class_name_or_spec_string: the first argument to the install script, if that's being run.
            Used to add subparsers for known Runnable classes or for custom Runnables defined at the time
            the Service/daemon is being installed.
        :type class_name_or_spec_string: str, optional

        :return: the :class:`~.workflow.OpenMSIStreamArgumentParser` object that should be used
        :rtype: :class:`~.workflow.OpenMSIStreamArgumentParser`

        :raises ValueError: if `install_or_manage` is neither "install" nor "manage"
        """
        parser = OpenMSIStreamArgumentParser()
        if install_or_manage=='install' :
            #subparsers from the classes that could be run
            subp_desc = 'The name of a "Runnable" class (or the specification for a custom Service code) to install '
            subp_desc = 'as a service must be given as the first argument. Adding the name (or spec string) of a class '
            subp_desc = 'that extends "Runnable" to the command line along with "-h" will show additional help.'
            if class_name_or_spec_string in ('--help','-h') :
                return parser
            parser.add_subparsers(description=subp_desc,required=True,dest='service_spec_string')
            if class_name_or_spec_string is None :
                for service_dict in SERVICE_CONST.available_services :
                    parser.add_subparser_arguments_from_class(service_dict['class'],addl_args=['optional_service_name'])
            elif class_name_or_spec_string in [d['class_name'] for d in SERVICE_CONST.available_services] :
                sds = [sd for sd in SERVICE_CONST.available_services if sd['class_name']==class_name_or_spec_string]
                s_d = sds[0]
                parser.add_subparser_arguments_from_class(s_d['class'],addl_args=['optional_service_name'])
            else :
                _, _, run_class, _ = cls.__parse_custom_service_string(class_name_or_spec_string)
                if run_class is not None :
                    parser.add_subparser_arguments_from_class(run_class,
                                                              subp_name=class_name_or_spec_string,
                                                              addl_args=['optional_service_name'])
                else :
                    parser.add_subparser_arguments(class_name_or_spec_string,['optional_service_name'])
        elif install_or_manage=='manage' :
            parser.add_arguments('service_name','run_mode','remove_env_vars','remove_install_args','remove_nssm')
        else :
            errmsg =  'ERROR: must call get_argument_parser with either "install" or "manage", '
            errmsg+= f'not "{install_or_manage}"!'
            raise ValueError(errmsg)
        return parser

    #################### PROPERTIES ####################

    @property
    def env_var_names(self) :
        """
        Names of the environment variables used by the service
        (any environment variables in the command line arguments or config file)
        """
        #get the names of environment variables from the command line and config file
        if self.argslist is not None and self.service_dict is not None :
            for arg in self.argslist :
                if arg.startswith('$') :
                    yield arg
            if self.service_dict['class'] is not None :
                parser = self.service_dict['class'].get_argument_parser()
                argsdests = [action.dest for action in parser.actions]
                if 'config' in argsdests :
                    pargs = parser.parse_args(args=self.argslist)
                    cfp = ConfigFileParser(pargs.config,logger=SERVICE_CONST.logger)
                    for evn in cfp.env_var_names :
                        yield evn



"""A DataFileStreamHandler that triggers some arbitrary local code when full files are available"""

#imports
from abc import ABC, abstractmethod
from ..config import RUN_OPT_CONST, DATA_FILE_HANDLING_CONST
from .data_file_chunk_handlers import DataFileChunkProcessor
from .data_file_stream_handler import DataFileStreamHandler
from .file_registry.stream_handler_registries import StreamProcessorRegistry

class DataFileStreamProcessor(DataFileStreamHandler,DataFileChunkProcessor,ABC) :
    """
    A class to consume :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk` messages into memory
    and perform some operation(s) when entire files are available.
    This is a base class that cannot be instantiated on its own.

    :param config_path: Path to the config file to use in defining the Broker connection and Consumers
    :type config_path: :class:`pathlib.Path`
    :param topic_name: Name of the topic to which the Consumers should be subscribed
    :type topic_name: str
    :param output_dir: Path to the directory where the log and csv registry files should be kept (if None a default
        will be created in the current directory)
    :type output_dir: :class:`pathlib.Path`, optional
    :param datafile_type: the type of data file that recognized files should be reconstructed as
        (must be a subclass of :class:`~.data_file_io.DownloadDataFileToMemory`)
    :type datafile_type: :class:`~.data_file_io.DownloadDataFileToMemory`, optional
    :param n_threads: the number of threads/consumers to run
    :type n_threads: int, optional
    :param consumer_group_id: the group ID under which each consumer should be created
    :type consumer_group_id: str, optional

    :raises ValueError: if `datafile_type` is not a subclass of
        :class:`~.data_file_io.DownloadDataFileToMemory`
    """

    def __init__(self,config_file,topic_name,**kwargs) :
        """
        Constructor method signature duplicated above to display in Sphinx docs
        """
        super().__init__(config_file,topic_name,**kwargs)

    def process_files_as_read(self) :
        """
        Consumes messages and stores their data in memory.
        Uses several parallel threads to consume message and calls :func:`~_process_downloaded_data_file`
        for fully read files. Runs until the user inputs a command to shut it down.

        :return: the total number of messages consumed
        :rtype: int
        :return: the total number of messages processed (registered in memory)
        :rtype: int
        :return: the paths of files successfully processed during the run
        :rtype: List
        """
        #startup message
        msg = f'Will process files from messages in the {self.topic_name} topic using {self.n_threads} '
        msg+= f'thread{"s" if self.n_threads>1 else ""}'
        self.logger.info(msg)
        #set up the stream processor registry
        self.file_registry = StreamProcessorRegistry(dirpath=self._output_dir,
                                                      topic_name=self.topic_name,
                                                      consumer_group_id=self.consumer_group_id,
                                                      logger=self.logger)
        #if there are files that need to be re-processed, set the variables to re-read messages from those files
        if self.file_registry.rerun_file_key_regex is not None :
            msg = f'Consumer{"s" if self.n_threads>1 else ""} will start from the beginning of the topic to '
            msg+= f're-read messages for {self.file_registry.n_files_to_rerun} previously-failed '
            msg+= f'file{"s" if self.file_registry.n_files_to_rerun>1 else ""}'
            self.logger.info(msg)
            self.restart_at_beginning=True
            self.message_key_regex=self.file_registry.rerun_file_key_regex
        #call the run loop
        self.run()
        #return the results of the processing
        return self.n_msgs_read, self.n_msgs_processed, self.completely_processed_filepaths

    def _process_message(self,lock,msg,rootdir_to_set=None):
        """
        Process a single message to add it to a file being held in memory until all messages are received.

        If the message failed to be decrypted, this method calls :func:`~_undecryptable_message_callback` and returns.

        If the message is the first one consumed for a particular file, or any message other than the last one needed,
        it registers the file as 'in_progress' in the .csv file.

        If the message is the last message needed for a file and its contents match the original hash
        of the file on disk, this method calls :func:`~_process_downloaded_data_file`.

        If the call to :func:`~_process_downloaded_data_file` returns None (success), this method moves the file to the
        'successfully processed' .csv file, and returns.

        If the call to :func:`~_process_downloaded_data_file` returns an Exception, this method calls
        :func:`~_failed_processing_callback`, registers the file as 'failed' in the .csv file, and returns.

        If the message is the last one needed but the contents are somehow different than the original file on disk,
        this method calls :func:`~_mismatched_hash_callback`, registers the file as 'mismatched_hash' in the .csv file,
        and returns.

        :param lock: Acquiring this :class:`threading.Lock` object ensures that only one instance
            of :func:`~_process_message` is running at once
        :type lock: :class:`threading.Lock`
        :param msg: The received :class:`confluent_kafka.KafkaMessage` object, or an undecrypted KafkaCrypto message
        :type msg: :class:`confluent_kafka.KafkaMessage` or :class:`kafkacrypto.Message`
        :param rootdir_to_set: Path to a directory that should be set as the "root" for reconstructed data files
            (default is the output directory)
        :type rootdir_to_set: :class:`pathlib.Path`

        :return: True if processing the message was successful (file in progress or successfully processed),
            False otherwise
        :rtype: bool
        """
        retval = super()._process_message(lock,msg,self._output_dir if rootdir_to_set is None else rootdir_to_set)
        #if the file was in progress or had a mismatched hash, return True or False, respectively
        if retval in (True,False) :
            return retval
        #get the DataFileChunk from the message value
        try :
            dfc = msg.value() #from a regular Kafka Consumer
        except TypeError :
            dfc = msg.value #from KafkaCrypto
        #if the file has had all of its messages read successfully, send it to the processing function
        if retval==DATA_FILE_HANDLING_CONST.FILE_SUCCESSFULLY_RECONSTRUCTED_CODE :
            if dfc.rootdir is not None :
                short_filepath = self.files_in_progress_by_path[dfc.filepath].full_filepath.relative_to(dfc.rootdir)
            else :
                short_filepath = self.files_in_progress_by_path[dfc.filepath].filepath
            self.logger.info(f'Processing {short_filepath}...')
            processing_retval = self._process_downloaded_data_file(self.files_in_progress_by_path[dfc.filepath],lock)
            to_return = None
            #if it was able to be processed
            if processing_retval is None :
                self.logger.info(f'Fully-read file {short_filepath} successfully processed')
                with lock :
                    self.file_registry.register_file_successfully_processed(dfc)
                    self.completely_processed_filepaths.append(dfc.filepath)
                to_return = True
            #warn if it wasn't processed correctly and invoke the callback
            else :
                if isinstance(processing_retval,Exception) :
                    errmsg = f'ERROR: Fully-read file {short_filepath} was not able to be processed. '
                    errmsg+= 'The traceback of the Exception thrown during processing will be logged below, but not '
                    errmsg+= 're-raised. The messages for this file will need to be consumed again if the file is to '
                    errmsg+= 'be processed! Please rerun with the same consumer ID to try again.'
                    self.logger.error(errmsg,exc_obj=processing_retval,reraise=False)
                else :
                    errmsg = f'Unrecognized return value from _process_downloaded_data_file: {processing_retval}'
                    self.logger.error(errmsg)
                with lock :
                    self.file_registry.register_file_processing_failed(dfc)
                self._failed_processing_callback(self.files_in_progress_by_path[dfc.filepath],lock)
                to_return = False
            #stop tracking the file
            with lock :
                del self.files_in_progress_by_path[dfc.filepath]
                del self.locks_by_fp[dfc.filepath]
            return to_return
        #otherwise the file is just in progress
        return True

    @abstractmethod
    def _process_downloaded_data_file(self,datafile,lock) :
        """
        Perform some arbitrary operation(s) on a given data file that has been fully read from the stream.
        Can optionally lock other threads using the given lock.

        Not implemented in the base class.

        :param datafile: A :class:`~.data_file_io.DownloadDataFileToMemory` object that has received
            all of its messages from the topic
        :type datafile: :class:`~.data_file_io.DownloadDataFileToMemory`
        :param lock: Acquiring this :class:`threading.Lock` object would ensure that only one instance
            of :func:`~_process_downloaded_data_file` is running at once
        :type lock: :class:`threading.Lock`

        :return: None if processing was successful, an Exception otherwise
        """
        raise NotImplementedError

    def _failed_processing_callback(self,datafile,lock) :
        """
        Called when :func:`~_process_downloaded_data_file` returns an Exception,
        providing an opportunity for fallback/backup processing in the case of failure.

        Does nothing in the base class.

        :param datafile: A :class:`~.data_file_io.DownloadDataFileToMemory` object that has received
            all of its messages from the topic
        :type datafile: :class:`~.data_file_io.DownloadDataFileToMemory`
        :param lock: Acquiring this :class:`threading.Lock` object would ensure that only one instance
            of :func:`~_failed_processing_callback` is running at once
        :type lock: :class:`threading.Lock`
        """

    def _on_check(self) :
        msg = f'{self.n_msgs_read} messages read, {self.n_msgs_processed} messages processed, '
        msg+= f'{len(self.completely_processed_filepaths)} files completely processed so far'
        self.logger.debug(msg)
        if len(self.files_in_progress_by_path)>0 or len(self.completely_processed_filepaths)>0 :
            self.logger.debug(self.progress_msg)

    @classmethod
    def get_command_line_arguments(cls):
        superargs,superkwargs = super().get_command_line_arguments()
        args = [*superargs,'topic_name']
        kwargs = {**superkwargs,'n_threads': RUN_OPT_CONST.N_DEFAULT_DOWNLOAD_THREADS}
        return args, kwargs



"""
A ControlledProcess running with more than one thread
"""

#imports
from threading import Lock
from abc import ABC, abstractmethod
from ..utilities.exception_tracking_thread import ExceptionTrackingThread
from .config import RUN_CONST
from .controlled_process import ControlledProcess

class ControlledProcessMultiThreaded(ControlledProcess,ABC) :
    """
    A class for running a group of processes in multiple threads until they're explicitly shut down.

    :param n_threads: the number of threads to use in the group of processes
    :type n_threads: int, optional
    """

    def __init__(self,*args,n_threads=RUN_CONST.DEFAULT_N_THREADS,**kwargs) :
        self.n_threads = n_threads
        super().__init__(*args,**kwargs)
        self.lock = Lock()
        self.__args_per_thread = []
        self.__kwargs_per_thread = {}
        self.__threads = []

    def run(self,args_per_thread=None,kwargs_per_thread=None) :
        """
        Create and start the set of independent threads running :func:`~_run_worker`, which can accept optional
        arguments/keyword arguments from the call to this function. Once the threads are started they will be
        automatically restarted if they crash.

        :param args_per_thread: a list of lists of arguments that should be given to the independent threads
            (one list of arguments per thread)
        :type args_per_thread: list, optional
        :param kwargs_per_thread: a list of dicts of keyword arguments that should be given to the independent threads
            (one dict of keyword arguments per thread)
        :type kwargs_per_thread: dict, optional
        """
        super().run()
        #correct the arguments for each thread
        if args_per_thread is not None :
            self.__args_per_thread = args_per_thread
        if self.__args_per_thread==[] or (not isinstance(self.__args_per_thread,list)) :
            self.__args_per_thread = [self.__args_per_thread]
        if not len(self.__args_per_thread)==self.n_threads :
            if not len(self.__args_per_thread)==1 :
                errmsg = 'ERROR: ControlledProcessMultiThreaded.run was given a list of arguments with '
                errmsg+= f'{len(self.__args_per_thread)} entries, but was set up to use {self.n_threads} threads!'
                self.logger.error(errmsg,ValueError)
            else :
                self.__args_per_thread = self.n_threads*self.__args_per_thread
        #correct the keyword arguments for each thread
        if kwargs_per_thread is not None :
            self.__kwargs_per_thread = kwargs_per_thread
        if not isinstance(self.__kwargs_per_thread,list) :
            self.__kwargs_per_thread = [self.__kwargs_per_thread]
        if not len(self.__kwargs_per_thread)==self.n_threads :
            if not len(self.__kwargs_per_thread)==1 :
                errmsg = 'ERROR: ControlledProcessMultiThreaded.run was given a list of arguments with '
                errmsg+= f'{len(self.__kwargs_per_thread)} entries, but was set up to use {self.n_threads} threads!'
                self.logger.error(errmsg,ValueError)
            else :
                self.__kwargs_per_thread = self.n_threads*self.__kwargs_per_thread
        #create and start the independent threads
        for i in range(self.n_threads) :
            self.__threads.append(ExceptionTrackingThread(target=self._run_worker,
                                           args=self.__args_per_thread[i],
                                           kwargs=self.__kwargs_per_thread[i]))
            self.__threads[-1].start()
        #loop while the process is alive, checking the control command queue and printing the "still alive" character
        while self.alive :
            self._print_still_alive()
            self._check_control_command_queue()
            self.__restart_crashed_threads()

    def _on_shutdown(self) :
        """
        Join all of the running threads. Can override this method further in subclasses, just be sure to also call
        :func:`super()._on_shutdown()`.
        """
        for thread in self.__threads :
            thread.join()

    @abstractmethod
    def _run_worker(self) :
        """
        The function that will actually be run in multiple threads while the process is alive. Should include a
        "while self.alive" loop to keep it running.

        `self.lock` can be used to guarantee exactly one copy of this function is running at a time.

        Not implemented in the base class.
        """
        raise NotImplementedError

    def __restart_crashed_threads(self) :
        """
        Log (but don't re-raise) Exceptions thrown by any of the threads and restart them based on the args/kwargs
        initially passed to :func:`~run`.
        """
        for ti,thread in enumerate(self.__threads) :
            if thread.caught_exception is not None :
                #log the error
                warnmsg = 'WARNING: a thread raised an Exception, which will be logged as an error below but not '
                warnmsg+= 'reraised. The thread that raised the error will be restarted.'
                self.logger.warning(warnmsg)
                self.logger.log_exception_as_error(thread.caught_exception,reraise=False)
                #try to join the thread
                try :
                    thread.join()
                except Exception :
                    pass
                finally :
                    self.__threads[ti] = None
                #restart the thread
                self.__threads[ti] = ExceptionTrackingThread(target=self._run_worker,
                                              args=self.__args_per_thread[ti],
                                              kwargs=self.__kwargs_per_thread[ti])
                self.__threads[ti].start()



"""A DataFile that will be broken into chunks and uploaded to a topic"""

#imports
import time
from threading import Thread
from queue import Queue
from hashlib import sha512
from ...kafka_wrapper import ProducerGroup
from ...workflow import Runnable
from ..config import RUN_OPT_CONST
from .. import DataFile
from .data_file_chunk import DataFileChunk

class UploadDataFile(DataFile,Runnable) :
    """
    Class to represent a data file whose messages will be uploaded to a topic

    Used as part of a :class:`~.DataFileUploadDirectory`, but can also be run standalone
    to upload a single file.

    :param filepath: The path to the file on disk
    :type filepath: :class:`pathlib.Path`
    :param to_upload: True (default) if the file should be uploaded. Used to set the existing files as
        "already uploaded" when running a :class:`~.DataFileUploadDirectory`.
    :type to_upload: bool, optional
    :param rootdir: path to the "root" directory that this file is in. Anything in the path beyond it
        will be added to the DataFileChunk so that it will be reconstructed inside a subdirectory
    :type rootdir: None or :class:`pathlib.Path`, optional
    :param filename_append: a string that should be appended to the end of the filename stem to distinguish the file
        that's produced from its original file on disk
    :type filename_append: None or str, optional
    """

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,filepath,to_upload=True,rootdir=None,filename_append=None,**kwargs) :
        """
        Constructor method
        """
        super().__init__(filepath,**kwargs)
        self.__to_upload = to_upload
        if rootdir is None :
            self.rootdir = self.filepath.parent
        else :
            self.rootdir = rootdir
        self.__filename_append = filename_append if filename_append is not None else ''
        self.__fully_enqueued = False
        self.__fully_produced = False
        self.chunks_to_upload = []
        self.__n_total_chunks = 0
        self.__chunk_infos = None
        self.__file_hash = None

    def upload_whole_file(self,config_path,topic_name,
                          n_threads=RUN_OPT_CONST.N_DEFAULT_UPLOAD_THREADS,
                          chunk_size=RUN_OPT_CONST.DEFAULT_CHUNK_SIZE) :
        """
        Chunk and upload an entire file on disk to a broker's topic.

        :param config_path: Path to the config file to use in defining the Broker connection and Producers
        :type config_path: :class:`pathlib.Path`
        :param topic_name: Name of the topic to which the file's messages should be produced
        :type topic_name: str
        :param n_threads: The number of threads/Producers to run at once during uploading
        :type n_threads: int, optional
        :param chunk_size: The size of the file chunk in each message in bytes
        :type chunk_size: int, optional
        """
        startup_msg = f"Uploading entire file {self.filepath} to {topic_name} in {chunk_size} byte chunks "
        startup_msg+=f"using {n_threads} threads...."
        self.logger.info(startup_msg)
        #add all the chunks to the upload queue
        upload_queue = Queue()
        self.enqueue_chunks_for_upload(upload_queue,chunk_size=chunk_size)
        #add "None" to the queue for each thread as the final values
        for _ in range(n_threads) :
            upload_queue.put(None)
        #produce all the messages in the queue using multiple threads
        producer_group = ProducerGroup(config_path,logger=self.logger)
        producers = []
        upload_threads = []
        for _ in range(n_threads) :
            producers.append(producer_group.get_new_producer())
            thread = Thread(target=producers[-1].produce_from_queue_looped, args=(upload_queue,topic_name))
            thread.start()
            upload_threads.append(thread)
        #join the threads
        for u_t in upload_threads :
            u_t.join()
        self.logger.info('Waiting for all enqueued messages to be delivered (this may take a moment)....')
        for producer in producers :
            producer.flush(timeout=-1) #don't leave the function until all messages have been sent/received
            producer.close()
        producer_group.close()
        self.logger.info('Done!')

    def add_chunks_to_upload(self,chunks_to_add=None,chunk_size=RUN_OPT_CONST.DEFAULT_CHUNK_SIZE) :
        """
        Add chunks from this file to the internal list of chunks to upload,
        possibly with some selection defined by :attr:`~select_bytes`

        :param chunks_to_add: a list of chunk indices to add to the list to be uploaded (Default=None adds all chunks)
        :type chunks_to_add: None or list[int], optional
        :param chunk_size: The size of the file chunk in each message in bytes
        :type chunk_size: int, optional
        """
        if self.__chunk_infos is None :
            try :
                self._build_list_of_file_chunks(chunk_size)
            except Exception as exc :
                fp = self.filepath.relative_to(self.rootdir) if self.rootdir is not None else self.filepath
                errmsg = f'ERROR: was not able to break {fp} into chunks for uploading. '
                errmsg+= 'Check log lines below for details on what went wrong. File will not be uploaded.'
                self.logger.error(errmsg,exc_obj=exc,reraise=False)
                self.__to_upload = False
                return
        #add the chunks to the final list as DataFileChunk objects
        for ichunk,chunk in enumerate(self.__chunk_infos,start=1) :
            if chunks_to_add is None or ichunk in chunks_to_add :
                self.chunks_to_upload.append(DataFileChunk(self.filepath,self.filename,self.__file_hash,
                                                            chunk[0],chunk[1],chunk[2],chunk[3],ichunk,
                                                            self.__n_total_chunks,
                                                            rootdir=self.rootdir,
                                                            filename_append=self.__filename_append))
        if len(self.chunks_to_upload)>0 and self.__fully_enqueued :
            self.__fully_enqueued = False

    def enqueue_chunks_for_upload(self,queue,n_threads=None,chunk_size=RUN_OPT_CONST.DEFAULT_CHUNK_SIZE,
                                   queue_full_timeout=0.001) :
        """
        Add some chunks of this file from the internal list to a given upload queue
        (the internal list will be created if :func:`~add_chunks_to_upload` hasn't already been called).

        When the entire internal list of file chunks has been added to the queue,
        the file will be marked as fully enqueued.

        If the given queue is full, this function will wait a bit and then return.

        :param queue: the Queue to which chunks should be added
        :type queue: :class:`queue.Queue`
        :param n_threads: the number of threads running during uploading; at most 5*this number of chunks will be
            added per call to this method. If this argument isn't given, every chunk in the internal list will be added.
        :type n_threads: None or int, optional
        :param chunk_size: The size of the file chunk in each message in bytes
        :type chunk_size: int, optional
        :param queue_full_timeout: amount of time to wait before returning if the queue is full
            and new messages can't be added
        :type queue_full_timeout: float, optional
        """
        if self.__fully_enqueued :
            warnmsg = f'WARNING: enqueue_chunks_for_upload called for fully enqueued file {self.filepath}, '
            warnmsg+= 'nothing else will be added.'
            self.logger.warning(warnmsg)
            return
        if queue.full() :
            time.sleep(queue_full_timeout)
            return
        if len(self.chunks_to_upload)==0 :
            self.add_chunks_to_upload(chunk_size=chunk_size)
        if n_threads is not None :
            n_chunks_to_add = 5*n_threads
        else :
            n_chunks_to_add = len(self.chunks_to_upload)
        ichunk = 0
        while len(self.chunks_to_upload)>0 and ichunk<n_chunks_to_add :
            while queue.full() :
                time.sleep(queue_full_timeout)
            queue.put(self.chunks_to_upload.pop(0))
            ichunk+=1
        if len(self.chunks_to_upload)==0 :
            self.__fully_enqueued = True

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _build_list_of_file_chunks(self,chunk_size) :
        """
        Build the list of DataFileChunks for this file

        chunk_size = the size of each chunk in bytes
        """
        #first make sure the choices of select_bytes are valid if necessary
        #and sort them by their start byte to keep the file hash in order
        if self.select_bytes :
            if not isinstance(self.select_bytes,list) :
                self.logger.error(f'ERROR: select_bytes={self.select_bytes} but is expected to be a list!',ValueError)
            for sbt in self.select_bytes :
                if (not isinstance(sbt,tuple)) or len(sbt)!=2 :
                    errmsg = f'ERROR: found {sbt} in select_bytes but all elements are expected to be two-entry tuples!'
                    self.logger.error(errmsg,ValueError)
                elif sbt[0]>=sbt[1] :
                    errmsg = f'ERROR: found {sbt} in select_bytes but start byte cannot be >= stop byte!'
                    self.logger.error(errmsg,ValueError)
            sorted_select_bytes = sorted(self.select_bytes,key=lambda x: x[0])
        #start a hash for the file and the lists of chunks
        file_hash = sha512()
        chunks = []
        isb = 0 #index for the current sorted_select_bytes entry if necessary
        #read the binary data in the file as chunks of the given size, adding each chunk to the list
        with open(self.filepath,'rb') as fp :
            chunk_offset = 0
            file_offset = 0 if (not self.select_bytes) else sorted_select_bytes[isb][0]
            n_bytes_to_read = chunk_size
            if self.select_bytes :
                n_bytes_to_read = min(chunk_size,sorted_select_bytes[isb][1]-file_offset)
            chunk = fp.read(n_bytes_to_read)
            while len(chunk) > 0 :
                file_hash.update(chunk)
                chunk_hash = sha512()
                chunk_hash.update(chunk)
                chunk_hash = chunk_hash.digest()
                chunk_length = len(chunk)
                chunks.append([chunk_hash,file_offset,chunk_offset,chunk_length])
                chunk_offset += chunk_length
                file_offset += chunk_length
                if self.select_bytes and file_offset==sorted_select_bytes[isb][1] :
                    isb+=1
                    if isb>(len(sorted_select_bytes)-1) :
                        break
                    file_offset=sorted_select_bytes[isb][0]
                n_bytes_to_read = chunk_size
                if self.select_bytes :
                    n_bytes_to_read = min(chunk_size,sorted_select_bytes[isb][1]-file_offset)
                fp.seek(file_offset)
                chunk = fp.read(n_bytes_to_read)
        file_hash = file_hash.digest()
        self.logger.info(f'File {self.filepath} has a total of {len(chunks)} chunks')
        #set the hash for the file
        self.__file_hash = file_hash
        #set the total number of chunks for this file
        self.__n_total_chunks = len(chunks)
        #build the list of all of the chunk infos for the file
        self.__chunk_infos = []
        for chunk in chunks :
            self.__chunk_infos.append(chunk)

    #################### CLASS METHODS ####################

    @classmethod
    def get_command_line_arguments(cls) :
        superargs,superkwargs = super().get_command_line_arguments()
        args = [*superargs,'filepath','config','topic_name','chunk_size']
        kwargs = {**superkwargs,'n_threads':RUN_OPT_CONST.N_DEFAULT_UPLOAD_THREADS}
        return args,kwargs

    @classmethod
    def run_from_command_line(cls,args=None) :
        """
        Run an :class:`~UploadDataFile` directly from the command line

        Calls :func:`~upload_whole_file` on an :class:`~.UploadDataFile` defined by command line (or given) arguments

        :param args: the list of arguments to send to the parser instead of getting them from sys.argv
        :type args: list
        """
        #make the argument parser
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        #make the DataFile for the single specified file
        upload_file = cls(args.filepath)
        #chunk and upload the file
        upload_file.upload_whole_file(args.config,args.topic_name,
                                      n_threads=args.n_threads,
                                      chunk_size=args.chunk_size)
        upload_file.logger.info(f'Done uploading {args.filepath}')

    #################### PROPERTIES ####################

    @property
    def select_bytes(self) :
        """
        In child classes, this property can be a list of tuples of (start_byte,stop_byte) in the file
        that will be the only ranges of bytes added when creating the list of chunks.
        The empty list in the base class will cause all bytes of the file to be uploaded.
        """
        return []
    @property
    def to_upload(self):
        """
        whether or not this file will be considered when uploading some group of data files
        """
        return self.__to_upload
    @to_upload.setter
    def to_upload(self,to_upload) :
        if to_upload and (not self.__to_upload) :
            self.logger.debug(f'Setting {self.filepath} to be uploaded')
        elif (not to_upload) and self.__to_upload :
            self.logger.debug(f'Setting {self.filepath} to NOT be uploaded')
        self.__to_upload = to_upload
    @property
    def fully_produced(self) :
        """
        True if this file has had all of its chunks successfully sent to the broker
        (only used when run as part of a :class:`~.DataFileUploadDirectory`)
        """
        return self.__fully_produced
    @fully_produced.setter
    def fully_produced(self,fp) :
        self.__fully_produced = fp
    @property
    def fully_enqueued(self) :
        """
        True if this file has had all of its chunks added to an upload queue
        """
        return self.__fully_enqueued
    @property
    def waiting_to_upload(self) :
        """
        True if this file is waiting for its upload to begin
        """
        if (not self.__to_upload) or self.__fully_enqueued :
            return False
        if len(self.chunks_to_upload)>0 :
            return False
        return True
    @property
    def upload_in_progress(self) :
        """
        True if this file is in the process of being enqueued to be uploaded
        """
        if (not self.__to_upload) or self.__fully_enqueued :
            return False
        if len(self.chunks_to_upload)==0 :
            return False
        return True
    @property
    def upload_status_msg(self) :
        """
        A message stating the file's name and status w.r.t. being enqueued to be uploaded
        """
        if self.rootdir is None :
            msg = f'{self.filepath} '
        else :
            msg = f'{self.filepath.relative_to(self.rootdir)} '
        if not self.__to_upload :
            msg+='(will not be uploaded)'
        elif self.__fully_produced :
            msg+=f'({self.__n_total_chunks} message'
            if self.__n_total_chunks!=1 :
                msg+='s'
            msg+=' fully produced to broker)'
        elif self.__fully_enqueued :
            msg+=f'({self.__n_total_chunks} message'
            if self.__n_total_chunks!=1 :
                msg+='s'
            msg+=' fully enqueued)'
        elif self.upload_in_progress :
            msg+=f'(in progress with {self.__n_total_chunks-len(self.chunks_to_upload)}'
            msg+=f'/{self.__n_total_chunks} messages enqueued)'
        elif self.waiting_to_upload :
            msg+=f'({self.__n_total_chunks} messages waiting to be enqueued)'
        else :
            msg+='(status unknown)'
        return msg

def main(args=None) :
    """
    main method to run from command line
    """
    UploadDataFile.run_from_command_line(args)

if __name__=='__main__' :
    main()



"""Wrapper around a Python ConfigParser to simplify some commonly-used operations"""

#imports
import os, configparser
from .logging import LogOwner

class ConfigFileParser(LogOwner) :
    """
    A class to parse configurations from files
    """

    #################### PROPERTIES ####################

    @property
    def has_default(self) :
        """
        True if a config file has a DEFAULT section
        """
        return 'DEFAULT' in self._config
    @property
    def available_group_names(self) :
        """
        The list of section names in the config file
        """
        return self._config.sections()
    @property
    def env_var_names(self) :
        """
        A generator of the environment variable names used in the config file
        """
        for csd in self._config.values() :
            for v in csd.values() :
                if v.startswith('$') :
                    yield v[1:]

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,config_path,*args,**kwargs) :
        """
        config_path = path to the config file to parse
        """
        super().__init__(*args,**kwargs)
        self.filepath = config_path
        if not config_path.is_file() :
            self.logger.error(f'ERROR: configuration file {config_path} does not exist!',FileNotFoundError)
        self._config = configparser.ConfigParser()
        self._config.read(config_path)

    def get_config_dict_for_groups(self,group_names) :
        """
        Return a config dictionary populated with configurations from groups with the given names

        group_names = the list of group names to add to the dictionary (or a single string)
        """
        if isinstance(group_names,str) :
            group_names = [group_names]
        config_dict = {}
        for group_name in group_names :
            if group_name not in self._config :
                self.logger.error(f'ERROR: {group_name} is not a recognized section in {self.filepath}!',ValueError)
            for key, value in self._config[group_name].items() :
                #don't add the 'node_id' to groups for brokers, producers, or consumers
                if key=='node_id' and group_name in ['broker','producer','consumer'] :
                    continue
                #if the value is an environment variable, expand it on the current system
                if value.startswith('$') :
                    exp_value = os.path.expandvars(value)
                    if exp_value == value :
                        errmsg = f'ERROR: Expanding {value} in {self.filepath} as an environment variable failed '
                        errmsg+= '(must be set on system)'
                        self.logger.error(errmsg,ValueError)
                    else :
                        value = exp_value
                config_dict[key] = value
        return config_dict

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _get_config_dict(self,group_name) :
        to_return = {}
        if group_name in self.available_group_names :
            to_return = self.get_config_dict_for_groups(group_name)
        return to_return



"""Classes for OpenMSIStream logging infrastructure"""

#imports
import pathlib, logging, traceback

class OpenMSIStreamFormatter(logging.Formatter) :
    """
    Very small extension of the usual logging.Formatter to allow modification of format based on message content
    """

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)

    def format(self, record):
        """
        If a message starts with a newline, start the actual logging line with the newline before any of the rest
        """
        formatted = ''
        if record.msg.startswith('\n') :
            record.msg = record.msg.lstrip('\n')
            formatted+='\n'
        formatted+=super().format(record)
        return formatted

class Logger :
    """
    Class for a general logger in OpenMSIStream format.

    :param logger_name: The name for the logger to use (automatically inferred from the running module if not given)
    :type logger_name: str, optional
    :param streamlevel: The level at/above which messages should be logged to the stream/console
    :type streamlevel: logging level int, optional
    :param logger_filepath: The path to a logger file to use or directory in which an automatically-named
        logger file should be created
    :type logger_filepath: :class:`pathlib.Path`, optional
    :param filelevel: The level at/above which messages should be written to the logfile
    :type filelevel: logging level int, optional
    """

    FORMATTER = OpenMSIStreamFormatter('[%(name)s at %(asctime)s] %(message)s','%Y-%m-%d %H:%M:%S')

    def __init__(self,logger_name=None,streamlevel=logging.DEBUG,logger_filepath=None,filelevel=logging.INFO) :
        """
        name = the name for this logger to use (probably something like the top module that owns it)
        """
        self._name = logger_name
        if self._name is None :
            self._name = self.__class__.__name__
        self._logger_obj = logging.getLogger(self._name)
        self._logger_obj.setLevel(logging.DEBUG)
        self._streamhandler = logging.StreamHandler()
        self._streamhandler.setLevel(streamlevel)
        self._streamhandler.setFormatter(self.FORMATTER)
        self._logger_obj.addHandler(self._streamhandler)
        self._filehandler = None
        if logger_filepath is not None :
            self.add_file_handler(logger_filepath,level=filelevel)

    def set_level(self,level) :
        """
        Set the level of the entire underlying logger

        :param level: The level to set
        :type level: logging level int
        """
        self._logger_obj.setLevel(level)

    def set_stream_level(self,level) :
        """
        Set the level of the underlying logger's streamhandler

        :param level: The level to set
        :type level: logging level int
        """
        self._streamhandler.setLevel(level)

    def set_file_level(self,level) :
        """
        Set the level of the underlying logger's filehandler

        :param level: The level to set
        :type level: logging level int
        """
        if self._filehandler is None :
            errmsg = f'ERROR: Logger {self._name} does not have a filehandler set but set_file_level was called!'
            raise RuntimeError(errmsg)
        self._filehandler.setLevel(level)

    def add_file_handler(self,filepath,level=logging.INFO) :
        """
        Add an additional :class:`logging.FileHandler` to the logger

        :param filepath: The path to the new logger file
        :type filepath: :class:`pathlib.Path`
        :param level: The level to set
        :type level: logging level int
        """
        if not isinstance(filepath,pathlib.PurePath) :
            self.error(f'ERROR: {filepath} is a {type(filepath)} object, not a Path object!',TypeError)
        if not filepath.is_file() :
            if not filepath.parent.is_dir() :
                filepath.parent.mkdir(parents=True)
            filepath.touch()
        self._filehandler = logging.FileHandler(filepath)
        self._filehandler.setLevel(level)
        self._filehandler.setFormatter(self.FORMATTER)
        self._logger_obj.addHandler(self._filehandler)

    #methods for logging different levels of messages

    def debug(self,msg,*args,**kwargs) :
        """
        Log a message at DEBUG level. Additional args/kwargs are sent to the underlying logger object's debug call.

        :param msg: the message to log
        :type msg: str
        """
        self._logger_obj.debug(msg,*args,**kwargs)

    def info(self,msg,*args,**kwargs) :
        """
        Log a message at INFO level. Additional args/kwargs are sent to the underlying logger object's info call.

        :param msg: the message to log
        :type msg: str
        """
        self._logger_obj.info(msg,*args,**kwargs)

    def warning(self,msg,*args,**kwargs) :
        """
        Log a message at WARNING level. Additional args/kwargs are sent to the underlying logger object's warning call.

        :param msg: the message to log (will have "WARNING: " prepended if it doesn't start with it)
        :type msg: str
        """
        if not msg.startswith('WARNING:') :
            msg = f'WARNING: {msg}'
        self._logger_obj.warning(msg,*args,**kwargs)

    #log an error message and optionally raise an exception with the same message, or raise a different exception
    def error(self,msg,exception_type=None,exc_obj=None,reraise=True,**kwargs) :
        """
        Log a message at ERROR level. Optionally raise an exception of a given type with the same message.
        Optionally log the traceback of a given Exception at ERROR level, reraising it afterward if desired.

        Additional kwargs are sent to the underlying logger object's error call(s).

        :param msg: the message to log
        :type msg: str
        :param exception_type: The type of Exception to raise with the same message as `msg`
        :type exception_type: :class:`BaseException`, optional
        :param exc_obj: An Exception object whose traceback should be logged at error level
        :type exc_obj: :class:`BaseException`, optional
        :param reraise: if True, the given `exc_obj` will be reraised after it's logged.
        :type reraise: bool
        """
        if not msg.startswith('ERROR:') :
            msg = f'ERROR: {msg}'
        self._logger_obj.error(msg,**kwargs)
        if exc_obj is not None :
            self.log_exception_as_error(exc_obj,reraise=reraise,**kwargs)
        if exception_type is not None :
            raise exception_type(msg)

    def log_exception_as_error(self,exc,reraise=True,**kwargs) :
        """
        Log the traceback of a given Exception as an error and optionally reraise it

        Additional kwargs are sent to the underlying logger object's error call.

        :param exc: An Exception object whose traceback should be logged at error level
        :type exc: :class:`BaseException`, optional
        :param reraise: if True, the given `exc` will be reraised after it's logged.
        :type reraise: bool
        """
        try :
            raise exc
        except Exception :
            exc_to_log = 'ERROR: Exception: ['
            for line in (traceback.format_exc()).split('\n') :
                exc_to_log+=f'{line},'
            exc_to_log+=']'
            self._logger_obj.error(exc_to_log,**kwargs)
            if reraise :
                raise exc

class LogOwner :
    """
    Any subclasses extending this one will have access to a Logger defined by the first class in the MRO to extend it

    :param logger: a :class:`~.utilities.Logger` object that this class should have access to. If this
        parameter is given it will override any of the others provided.
    :type logger: :class:`~.utilities.Logger`, optional
    :param logger_name: The name for the logger to use (automatically inferred from the running module if not given)
    :type logger_name: str, optional
    :param streamlevel: The level at/above which messages should be logged to the stream/console
    :type streamlevel: logging level int, optional
    :param logger_filepath: The path to a logger file to use or directory in which an automatically-named
        logger file should be created
    :type logger_filepath: :class:`pathlib.Path`, optional
    :param filelevel: The level at/above which messages should be written to the logfile
    :type filelevel: logging level int, optional
    """

    @property
    def logger(self) :
        """
        The logger object that the class can use
        """
        return self.__logger

    @logger.setter
    def logger(self,logger) :
        if ( hasattr(self,'_LogOwner__logger') and
             self.__logger is not None and
             (not isinstance(self.__logger,type(logger))) ) :
            errmsg = f'ERROR: tried to reset a logger of type {type(self.__logger)} to a new logger of type {logger}!'
            self.__logger.error(errmsg,ValueError)
        else :
            self.__logger = logger

    def __init__(self,*args,
                 logger=None,logger_name=None,streamlevel=logging.DEBUG,logger_file=None,filelevel=logging.INFO,
                 **other_kwargs) :
        if logger is not None :
            self.__logger = logger
        else :
            if logger_name is None :
                logger_name = self.__class__.__name__
            logger_filepath = logger_file
            if logger_file is not None and logger_file.is_dir() :
                logger_filepath = logger_file / f'{self.__class__.__name__}.log'
            self.__logger = Logger(logger_name,streamlevel,logger_filepath,filelevel)
        super().__init__(*args,**other_kwargs)



"""Custom argument parser and associated functions"""

#imports
import pathlib, math, re
from argparse import ArgumentParser
from ..data_file_io.config import RUN_OPT_CONST
from .config import RUN_CONST

#################### MISC. FUNCTIONS ####################

def existing_file(argstring) :
    """
    convert a string or path argument into a path to a file, checking if it exists
    """
    filepath = pathlib.Path(argstring)
    if filepath.is_file() :
        return filepath.resolve()
    raise FileNotFoundError(f'ERROR: file {argstring} does not exist!')

def existing_dir(argstring) :
    """
    convert a string or path argument into a directory path, checking if it exists
    """
    dirpath = pathlib.Path(argstring)
    if dirpath.is_dir() :
        return dirpath.resolve()
    raise FileNotFoundError(f'ERROR: directory {argstring} does not exist!')

def create_dir(argstring) :
    """
    convert a string or path argument into a directory path, creating it if necessary
    """
    if argstring is None : #Then the argument wasn't given and nothing should be done
        return None
    dirpath = pathlib.Path(argstring)
    if dirpath.is_dir() :
        return dirpath.resolve()
    if dirpath.exists() :
        raise RuntimeError(f'ERROR: directory path {argstring} exists but is not a directory!')
    dirpath.mkdir(parents=True)
    return dirpath.resolve()

def config_path(configarg) :
    """
    convert a string or path argument into a config file path (raise an exception if the file can't be found)
    """
    if isinstance(configarg,str) and '.' not in configarg :
        configarg+=RUN_CONST.CONFIG_FILE_EXT
    configpath = pathlib.Path(configarg)
    if configpath.is_file() :
        return configpath.resolve()
    if (RUN_CONST.CONFIG_FILE_DIR/configpath).is_file() :
        return (RUN_CONST.CONFIG_FILE_DIR/configpath).resolve()
    raise ValueError(f'ERROR: config argument {configarg} is not a recognized config file!')

def detect_bucket_name(argstring) :
    """
    detects if the bucket name contains invalid characters
    """
    if argstring is None : #Then the argument wasn't given and nothing should be done
        return None
    illegal_charcters = ['#', '%', '&' ,'{', '}', '\\', '/', '<', '>', '*', '?', ' ',
                         '$', '!', '\'', '\"', ':', '@', '+', '`', '|', '=']
    if argstring in illegal_charcters:
        raise RuntimeError(f'ERROR: Illegal characters in bucket_name {argstring}')
    return argstring

def int_power_of_two(argval) :
    """
    make sure a given value is a nonzero integer power of two (or can be converted to one)
    """
    if not isinstance(argval,int) :
        argval=int(argval)
    if argval<=0 or math.ceil(math.log2(argval))!=math.floor(math.log2(argval)) :
        raise ValueError(f'ERROR: invalid argument: {argval} must be a (nonzero) power of two!')
    return argval

def positive_int(argval) :
    """
    make sure a given value is a positive integer
    """
    argval = int(argval)
    if (not isinstance(argval,int)) or (argval<1) :
        raise ValueError(f'ERROR: invalid argument: {argval} must be a positive integer!')
    return argval

#################### MYARGUMENTPARSER CLASS ####################

class OpenMSIStreamArgumentParser(ArgumentParser) :
    """
    Class to make it easier to get an ArgumentParser with some commonly-used arguments in it.

    All constructor arguments get passed to the underlying :class:`argparse.ArgumentParser` object.

    Arguments for the parser are defined in the :attr:`~OpenMSIStreamArgumentParser.ARGUMENTS` class variable,
    which is a dictionary. The keys are names of arguments, and the values are lists. The first entry in each list
    is a string reading "positional" or "optional" depending on the type of argument, and the second entry is a
    dictionary of keyword arguments to send to :func:`argparse.ArgumentParser.add_argument`.
    """

    ARGUMENTS = {
        'filepath':
            ['positional',{'type':existing_file,'help':'Path to the data file to use'}],
        'output_dir':
            ['positional',{'type':create_dir,'help':'Path to the directory to put output in'}],
        'upload_dir':
            ['positional',{'type':existing_dir,'help':'Path to the directory to watch for files to upload'}],
        'bucket_name':
            ['positional', {'type': detect_bucket_name,
                            'help': '''Name of the object store bucket to which consumed files should be transferred.
                                       Name should only contain valid characters'''}],
        'config':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_CONFIG_FILE,'type':config_path,
                         'help':f'''Name of config file to use in {RUN_CONST.CONFIG_FILE_DIR.resolve()},
                                    or path to a file in a different location'''}],
        'topic_name':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_TOPIC_NAME,
                         'help':'Name of the topic to produce to or consume from'}],
        'consumer_topic_name':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_TOPIC_NAME,
                         'help':'Name of the topic to consume from'}],
        'producer_topic_name':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_TOPIC_NAME,
                         'help':'Name of the topic to produce to'}],
        'n_threads':
            ['optional',{'default':RUN_CONST.DEFAULT_N_THREADS,'type':positive_int,
                         'help':'Maximum number of threads to use'}],
        'n_consumer_threads':
            ['optional',{'default':RUN_OPT_CONST.N_DEFAULT_DOWNLOAD_THREADS,'type':positive_int,
                         'help':'Number of consumer threads to use'}],
        'n_producer_threads':
            ['optional',{'default':RUN_OPT_CONST.N_DEFAULT_UPLOAD_THREADS,'type':positive_int,
                         'help':'Number of producer threads to use'}],
        'upload_regex':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_UPLOAD_REGEX,'type':re.compile,
                         'help':'Only files matching this regular expression will be uploaded'}],
        'chunk_size':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_CHUNK_SIZE,'type':int_power_of_two,
                         'help':'''Max size (in bytes) of chunks into which files should be broken
                                   as they are uploaded'''}],
        'queue_max_size':
            ['optional',{'default':RUN_OPT_CONST.DEFAULT_MAX_UPLOAD_QUEUE_SIZE,'type':int,
                         'help':'''Maximum number of items to allow in the upload queue at a time.
                                 Use to limit RAM usage or throttle production rate if necessary.'''}],
        'update_seconds':
            ['optional',{'default':RUN_CONST.DEFAULT_UPDATE_SECONDS,'type':int,
                         'help':'''Number of seconds between printing a "." to the console
                                   to indicate the program is alive'''}],
        'upload_existing':
            ['optional',{'action':'store_true',
                         'help':'''Add this flag to upload files already existing in addition to those
                                   added to the directory after this code starts running (by default
                                   only files added after startup will be uploaded)'''}],
        'consumer_group_id':
            ['optional',{'default':'create_new',
                         'help':'ID to use for all consumers in the group'}],
        'optional_output_dir':
            ['optional',{'type':create_dir,
                         'help':'Optional path to directory to put output in'}],
        'service_name':
            ['positional',{'help':'The name of the service to work with'}],
        'optional_service_name':
            ['optional',{'help':'The customized name of the Service that will be installed to run the chosen class'}],
        'run_mode':
            ['positional',{'choices':['start','status','stop','remove','stop_and_remove',
                                      'reinstall','stop_and_reinstall','stop_and_restart'],
                           'help':'What to do with the service'}],
        'remove_env_vars':
            ['optional',{'action':'store_true',
                         'help':'''Add this flag to also remove username/password environment variables
                                   when removing a Service'''}],
        'remove_install_args':
            ['optional',{'action':'store_true',
                         'help':'Add this flag to also remove the install arguments file when removing a Service'}],
        'remove_nssm':
            ['optional',{'action':'store_true',
                         'help':'Add this flag to also remove the NSSM executable when removing a Service'}],
    }

    #################### OVERLOADED FUNCTIONS ####################

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)
        self.__argnames_added = []
        self.__subparsers_action_obj = None
        self.__subparsers = {}
        self.__subparser_argnames_added = {}

    def add_subparsers(self,*args,**kwargs) :
        """
        Overloaded from base class; OpenMSIStreamArgumentParser actually owns its subparsers to simplify conflicting
        argument names
        """
        if self.__subparsers_action_obj is not None or self.__subparsers :
            raise RuntimeError('ERROR: add_subparsers called for an argument parser that already has subparsers!')
        self.__subparsers_action_obj = super().add_subparsers(*args,**kwargs)

    def parse_args(self,*args,**kwargs) :
        """
        Overloaded from base class to print usage when catching errors in parsing arguments
        """
        try :
            return super().parse_args(*args,**kwargs)
        except Exception :
            self.print_usage()
            raise

    #################### UNIQUE FUNCTIONS ####################

    def add_arguments(self,*args,**kwargs) :
        """
        Add a group of common arguments to the parser

        :param args: Names of arguments that should be added exactly as they're listed in
            :attr:`~OpenMSIStreamArgumentParser.ARGUMENTS`
        :type args: list
        :param kwargs: Dictionary whose keys are argument names like in `args` and whose values are new
            default argument values
        :type kwargs: dict

        :raises ValueError: if the name of an argument isn't recognized in
            :attr:`~OpenMSIStreamArgumentParser.ARGUMENTS`, or if the type of a new default argument is different
            than the type of its original default argument
        """
        if len(args)<1 and len(kwargs)<1 :
            raise ValueError('ERROR: must specify at least one desired argument to create an argument parser!')
        args_to_use = []
        for argname in args :
            if argname not in args_to_use :
                args_to_use.append(argname)
        for argname in kwargs :
            if argname in args_to_use :
                args_to_use.remove(argname)
        for argname in args_to_use :
            argname_to_add, kwargs_for_arg = self.__get_argname_and_kwargs(argname)
            if argname_to_add not in self.__argnames_added :
                self.add_argument(argname_to_add,**kwargs_for_arg)
                self.__argnames_added.append(argname_to_add)
        for argname,argdefault in kwargs.items() :
            argname_to_add, kwargs_for_arg = self.__get_argname_and_kwargs(argname,argdefault)
            if argname_to_add not in self.__argnames_added :
                self.add_argument(argname_to_add,**kwargs_for_arg)
                self.__argnames_added.append(argname_to_add)

    def add_subparser_arguments(self,subp_name,args_to_add=None,kwargs_to_add=None,**other_kwargs) :
        """
        Create a new subparser and add arguments to it.

        :param subp_name: the name of the subparser command
        :type subp_name: str
        :param args_to_add: additional arguments that should be added to the subparser
            (same format as `args` for :func:`~OpenMSIStreamArgumentParser.add_arguments`)
        :type args_to_add: list, optional
        :param kwargs_to_add: additional keyword arguments that should be added to the subparser
            (same format as `kwargs` for :func:`~OpenMSIStreamArgumentParser.add_arguments`)
        :type kwargs_to_add: dict, optional
        :param other_kwargs: any other keyword arguments are passed to the subparser's add_parser() method
        :type other_kwargs: dict, optional

        :raises ValueError: like in :func:`~OpenMSIStreamArgumentParser.add_arguments`
        """
        if self.__subparsers_action_obj is None :
            errmsg = 'ERROR: add_subparser_arguments called for an argument parser that '
            errmsg+= 'has not added subparsers!'
            raise RuntimeError(errmsg)
        if subp_name in self.__subparsers :
            errmsg = f'ERROR: subparser arguments for {subp_name} have already been added to this argument parser!'
            raise RuntimeError(errmsg)
        self.__subparsers[subp_name] = self.__subparsers_action_obj.add_parser(subp_name,**other_kwargs)
        self.__subparser_argnames_added[subp_name] = []
        if args_to_add is not None :
            for argname in args_to_add :
                argname_to_add, kwargs_for_arg = self.__get_argname_and_kwargs(argname)
                if argname_to_add not in self.__subparser_argnames_added[subp_name] :
                    self.__subparsers[subp_name].add_argument(argname_to_add,**kwargs_for_arg)
                    self.__subparser_argnames_added[subp_name].append(argname_to_add)
        if kwargs_to_add is not None :
            for argname,argdefault in kwargs_to_add.items() :
                argname_to_add,kwargs_for_arg = self.__get_argname_and_kwargs(argname,argdefault)
                if argname_to_add not in self.__subparser_argnames_added[subp_name] :
                    self.__subparsers[subp_name].add_argument(argname_to_add,**kwargs_for_arg)
                    self.__subparser_argnames_added[subp_name].append(argname_to_add)

    def add_subparser_arguments_from_class(self,class_to_add,*,
                                           subp_name=None,addl_args=None,addl_kwargs=None,**other_kwargs) :
        """
        Create a new subparser and add arguments from the given class to it.
        `class_to_add must` inherit from :class:`~Runnable` to be able to get its arguments.

        :param subp_name: an override for the name of the command the subparser should be registered under.
            (Default is the name of the class.)
        :type subp_name: str, optional
        :param addl_args: additional arguments that should be added to the subparser
            (same format as `args` for :func:`~OpenMSIStreamArgumentParser.add_arguments`)
        :type addl_args: list, optional
        :param addl_kwargs: additional keyword arguments that should be added to the subparser
            (same format as `kwargs` for :func:`~OpenMSIStreamArgumentParser.add_arguments`)
        :type addl_kwargs: dict, optional
        :param other_kwargs: any other keyword arguments are passed to subparsers' add_parser() method
        :type other_kwargs: dict, optional

        :raises ValueError: like in :func:`~OpenMSIStreamArgumentParser.add_arguments`
        """
        if subp_name is None :
            subp_name = class_to_add.__name__
        argnames, argnames_with_defaults = class_to_add.get_command_line_arguments()
        if addl_args is not None :
            argnames = [*argnames,*addl_args]
        if addl_kwargs is not None :
            argnames_with_defaults = {**argnames_with_defaults,**addl_kwargs}
        self.add_subparser_arguments(subp_name,argnames,argnames_with_defaults,**other_kwargs)

    def __get_argname_and_kwargs(self,argname,new_default=None) :
        """
        Return the name and kwargs dict for a particular argument

        argname = the given name of the argument, will be matched to a name in ARGUMENTS
        new_default = an optional parameter given to override the default already specified in ARGUMENTS
        """
        if argname in self.ARGUMENTS :
            if self.ARGUMENTS[argname][0]=='positional' :
                argname_to_add = argname
            else :
                if argname.startswith('optional_') :
                    argname_to_add = f'--{argname[len("optional_"):]}'
                else :
                    argname_to_add = f'--{argname}'
            kwargs = self.ARGUMENTS[argname][1].copy()
            if new_default is not None :
                if 'default' in kwargs.keys() and not isinstance(new_default,type(kwargs['default'])) :
                    errmsg = f'ERROR: new default value {new_default} for argument {argname} is of a different type '
                    errmsg+= f'than expected based on the old default ({self.ARGUMENTS[argname]["kwargs"]["default"]})!'
                    raise ValueError(errmsg)
                kwargs['default'] = new_default
            if 'default' in kwargs.keys() :
                if 'help' in kwargs.keys() :
                    kwargs['help']+=f" (default = {kwargs['default']})"
                else :
                    kwargs['help']=f"default = {kwargs['default']}"
            return argname_to_add, kwargs
        raise ValueError(f'ERROR: argument {argname} is not recognized as an option!')

    @property
    def actions(self) :
        """
        wrapper around parser._actions to make it publicly available
        """
        return self._actions



"""
Code for managing the two csv files listing files that are
in the process of being produced/have been fully produced
"""

#imports
import pathlib, datetime
from typing import Set
from dataclasses import dataclass
from ....utilities import DataclassTable, LogOwner

@dataclass
class RegistryLineInProgress :
    """
    A line in the table listing files in progress
    """
    filename : str
    filepath : pathlib.Path
    n_chunks : int
    n_chunks_delivered : int
    n_chunks_to_send : int
    started : datetime.datetime
    chunks_delivered : Set[int]
    chunks_to_send : Set[int]

@dataclass
class RegistryLineCompleted :
    """
    A line in the table listing completed files
    """
    filename : str
    filepath : pathlib.Path
    n_chunks : int
    started : datetime.datetime
    completed : datetime.datetime

class ProducerFileRegistry(LogOwner) :
    """
    A class to manage two atomic csv files listing which portions
    of which files have been uploaded to a particular topic
    """

    def __init__(self,dirpath,topic_name,*args,**kwargs) :
        """
        dirpath    = path to the directory that should contain the csv file
        topic_name = the name of the topic that will be produced to (used in the filename)
        """
        super().__init__(*args,**kwargs)
        #A table to list the files currently in progress
        in_prog_filepath = dirpath / f'files_to_upload_to_{topic_name}.csv'
        self.__in_prog = DataclassTable(dataclass_type=RegistryLineInProgress,filepath=in_prog_filepath,
                                        logger=self.logger)
        #A table to list the files that have been completely produced
        completed_filepath = dirpath / f'files_fully_uploaded_to_{topic_name}.csv'
        self.__completed = DataclassTable(dataclass_type=RegistryLineCompleted,filepath=completed_filepath,
                                          logger=self.logger)

    def get_incomplete_filepaths_and_chunks(self) :
        """
        Generate tuples of (filepath, chunks to upload) for each file that has not yet been completely uploaded
        """
        for obj_address in self.__in_prog.obj_addresses :
            attr_dict = self.__in_prog.get_entry_attrs(obj_address,'filepath','chunks_to_send')
            if len(attr_dict['chunks_to_send'])>0 :
                yield attr_dict['filepath'],attr_dict['chunks_to_send']

    def get_completed_filepaths(self) :
        """
        Generate filepaths for each file that has been completely uploaded
        """
        for obj_address in self.__completed.obj_addresses :
            yield self.__completed.get_entry_attrs(obj_address,'filepath')

    def register_chunk(self,filename,filepath,n_total_chunks,chunk_i) :
        """
        Register a chunk as having been successfully produced
        Returns True if all chunks for the file have been produced
        """
        #get a dictionary of the existing object addresses keyed by their filepaths
        existing_obj_addresses = self.__in_prog.obj_addresses_by_key_attr('filepath')
        #if the file is already recognized as in progress
        if filepath in existing_obj_addresses :
            return self.__register_chunk_of_existing_file(filename,filepath,n_total_chunks,chunk_i)
        #otherwise it's a new file to list somewhere
        return self.__register_chunk_for_new_file(filename,filepath,n_total_chunks,chunk_i)

    def __register_chunk_of_existing_file(self,filename,filepath,n_total_chunks,chunk_i) :
        """
        Register a chunk that belongs to an existing file.
        Returns True if all chunks for the file have been received by the broker.
        """
        #get a dictionary of the existing object addresses keyed by their filepaths
        existing_obj_addresses = self.__in_prog.obj_addresses_by_key_attr('filepath')
        #make sure there's only one object with this filepath
        if len(existing_obj_addresses[filepath])!=1 :
            errmsg = f'ERROR: found {len(existing_obj_addresses[filepath])} files in the producer registry '
            errmsg+= f'for filepath {filepath}'
            self.logger.error(errmsg,RuntimeError)
        existing_addr = existing_obj_addresses[filepath][0]
        #make sure the total numbers of chunks match
        existing_n_chunks = self.__in_prog.get_entry_attrs(existing_addr,'n_chunks')
        if existing_n_chunks!=n_total_chunks :
            errmsg = f'ERROR: {filepath} in {self.__class__.__name__} is listed as having '
            errmsg+= f'{existing_n_chunks} total chunks, but the producer callback for this file '
            errmsg+= f'lists {n_total_chunks} chunks! Did the chunk size change?'
            self.logger.error(errmsg,RuntimeError)
        #editing the object's attributes has to be done in one thread at a time
        with self.__in_prog.lock :
            #get its current state
            attrs = self.__in_prog.get_entry_attrs(existing_addr,'chunks_delivered','chunks_to_send')
            #if the chunk is already registered, just return
            if chunk_i in attrs['chunks_delivered'] :
                return False
            #otherwise add/remove it from the sets
            attrs['chunks_delivered'].add(chunk_i)
            try :
                attrs['chunks_to_send'].remove(chunk_i)
            except KeyError :
                pass
            #reset the attributes for the object
            new_attrs = {
                'n_chunks_delivered':len(attrs['chunks_delivered']),
                'n_chunks_to_send':len(attrs['chunks_to_send']),
                'chunks_delivered':attrs['chunks_delivered'],
                'chunks_to_send':attrs['chunks_to_send'],
                }
        self.__in_prog.set_entry_attrs(existing_addr,**new_attrs)
        # if there are no more chunks to send and all chunks have been delivered,
        # move this file from "in progress" to "completed" and force-dump the files
        if ( self.__in_prog.get_entry_attrs(existing_addr,'n_chunks_delivered')==n_total_chunks and
                self.__in_prog.get_entry_attrs(existing_addr,'n_chunks_to_send')==0 ) :
            started = self.__in_prog.get_entry_attrs(existing_addr,'started')
            completed_entry = RegistryLineCompleted(filename,filepath,n_total_chunks,
                                                    started,datetime.datetime.now())
            self.__completed.add_entries(completed_entry)
            self.__completed.dump_to_file()
            self.__in_prog.remove_entries(existing_addr)
            self.__in_prog.dump_to_file()
            return True
        return False

    def __register_chunk_for_new_file(self,filename,filepath,n_total_chunks,chunk_i) :
        """
        Register the first chunk for a file.
        Returns True if the file only had one chunk.
        """
        to_deliver = set([])
        for ichunk in range(1,n_total_chunks+1) :
            if ichunk!=chunk_i :
                to_deliver.add(ichunk)
        #if there are other chunks to deliver, register it to "in_progress"
        if len(to_deliver)>0 :
            in_prog_entry = RegistryLineInProgress(filename,filepath,n_total_chunks,
                                                    1,n_total_chunks-1,datetime.datetime.now(),
                                                    set([chunk_i]),to_deliver)
            self.__in_prog.add_entries(in_prog_entry)
            self.__in_prog.dump_to_file()
            return False
        #otherwise register it as a completed file
        completed_entry = RegistryLineCompleted(filename,filepath,n_total_chunks,
                                                datetime.datetime.now(),datetime.datetime.now())
        self.__completed.add_entries(completed_entry)
        self.__completed.dump_to_file()
        return True
